from functools import cached_property
from typing import Literal, Optional, Union

import h5py
import pandas as pd
import zarr
from anndata import AnnData
from anndata._io.specs.methods import _read_partial
from anndata._io.specs.registry import read_elem, read_elem_partial
from lndb.dev.upath import infer_filesystem as _infer_filesystem
from lnschema_core import File
from lnschema_core.dev._storage import filepath_from_file

from ._lazy_field import LazySelector


def _indices(base_indices, select_indices):
    if len(base_indices) == len(select_indices):
        return slice(None)
    else:
        return list(base_indices.get_indexer(select_indices))


def _subset_adata_storage(
    storage: Union[zarr.Group, h5py.File],
    query_obs: Optional[Union[str, LazySelector]] = None,
    query_var: Optional[Union[str, LazySelector]] = None,
) -> Union[AnnData, None]:
    with storage as access:
        obs = read_elem(access["obs"])
        var = read_elem(access["var"])

        if query_obs is not None:
            if hasattr(query_obs, "evaluate"):
                obs_result = obs[query_obs.evaluate(obj=obs)]  # type: ignore
            else:
                obs_result = obs.query(query_obs)
        else:
            obs_result = obs

        if query_var is not None:
            if hasattr(query_var, "evaluate"):
                var_result = var[query_var.evaluate(obj=var)]  # type: ignore
            else:
                var_result = var.query(query_var)
        else:
            var_result = var

        if obs_result.index.empty or var_result.index.empty:
            return None

        obs_idx = _indices(obs.index, obs_result.index)
        var_idx = _indices(var.index, var_result.index)

        prepare_adata = {}
        prepare_adata["obs"] = obs_result
        prepare_adata["var"] = var_result
        X = read_elem_partial(access["X"], indices=(obs_idx, var_idx))
        prepare_adata["X"] = X
        if "obsm" in access:
            obsm = _read_partial(access["obsm"], indices=(obs_idx, slice(None)))
            prepare_adata["obsm"] = obsm
        if "varm" in access:
            varm = _read_partial(access["varm"], indices=(var_idx, slice(None)))
            prepare_adata["varm"] = varm
        if "obsp" in access:
            obsp = _read_partial(access["obsp"], indices=(obs_idx, obs_idx))
            prepare_adata["obsp"] = obsp
        if "varp" in access:
            varp = _read_partial(access["varp"], indices=(var_idx, var_idx))
            prepare_adata["varp"] = varp
        if "layers" in access:
            layers = _read_partial(access["layers"], indices=(obs_idx, var_idx))
            prepare_adata["layers"] = layers
        if "uns" in access:
            prepare_adata["uns"] = read_elem(access["uns"])

        return AnnData(**prepare_adata)


def _subset_anndata_file(
    file: File,
    query_obs: Optional[Union[str, LazySelector]] = None,
    query_var: Optional[Union[str, LazySelector]] = None,
) -> Union[AnnData, None]:
    file_path = filepath_from_file(file)
    fs, file_path_str = _infer_filesystem(file_path)

    adata: Union[AnnData, None] = None

    if file.suffix == ".h5ad":
        with fs.open(file_path_str, mode="rb") as file:
            storage = h5py.File(file, mode="r")
            adata = _subset_adata_storage(storage, query_obs, query_var)
    elif file.suffix == ".zarr":
        mapper = fs.get_mapper(file_path_str, check=True)
        storage = zarr.open(mapper, mode="r")
        adata = _subset_adata_storage(storage, query_obs, query_var)

    return adata


def get_obs_var_storage(
    storage: Union[zarr.Group, h5py.File], which=Literal["obs", "var"]
) -> pd.DataFrame:
    with storage as access:
        result = read_elem(access[which])
    if isinstance(result, pd.DataFrame):
        return result
    # is array
    elif "index" in result.dtype.fields:
        return pd.DataFrame.from_records(result, index="index")
    else:
        return pd.DataFrame.from_records(result)


class CloudAnnData:
    def __init__(self, file: File):
        self._file = file

    def _get_obs_var(self, which=Literal["obs", "var"]) -> pd.DataFrame:
        file_path = filepath_from_file(self._file)
        fs, file_path_str = _infer_filesystem(file_path)
        if self._file.suffix == ".h5ad":
            with fs.open(file_path_str, mode="rb") as f:
                storage = h5py.File(f, mode="r")
                df = get_obs_var_storage(storage, which)
        elif self._file.suffix == ".zarr" or self._file.suffix == ".zrad":
            mapper = fs.get_mapper(file_path_str, check=True)
            storage = zarr.open(mapper, mode="r")
            df = get_obs_var_storage(storage, which)
        return df

    @cached_property
    def obs(self) -> pd.DataFrame:
        return self._get_obs_var(which="obs")

    @cached_property
    def var(self) -> pd.DataFrame:
        return self._get_obs_var(which="var")



from typing import List

from lnschema_core import File, Folder

from .dev.db._select import select

Folder.__doc__ = """Folders: collections of files.

- Guide: :doc:`/guide/folder`

Real vs. virtual folders:

- A real LaminDB `Folder` has a 1:1 correspondence to a folder on a file system
  or in object storage, and a `.key` that is not `None`.
- A virtual LaminDB `Folder` is a mere way of grouping files. A file can be
  linked to multiple virtual folders, but only to one real folder.
"""


# exposed to users as Folder.subset()
def subset(self: Folder, *, prefix: str, **fields) -> List[File]:
    """Get files via relative path to folder."""
    # ensure is actual folder, not virtual one
    if self.key is None:
        raise ValueError(
            ".get() is only defined for real folders, not virtual ones"
            "you can access files via .files or by refining with queries"
        )
    files = (
        select(File, **fields).where(File.key.startswith(self.key + "/" + prefix)).all()
    )
    return files


Folder.subset = subset



"""v0.5.0.

Revision ID: 3badf20f18c8
Revises: 01fcb82dafd4
Create Date: 2022-08-26 14:17:51.167740
"""
import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "3badf20f18c8"
down_revision = "01fcb82dafd4"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "pipeline",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("reference", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.PrimaryKeyConstraint("id", "v"),
    )
    op.create_index(op.f("ix_pipeline_name"), "pipeline", ["name"], unique=False)
    op.create_index(op.f("ix_pipeline_reference"), "pipeline", ["reference"], unique=False)
    op.add_column("dobject", sa.Column("storage_id", sqlmodel.sql.sqltypes.AutoString()))
    op.alter_column("dobject", "dtransform_id", existing_type=sa.VARCHAR())
    op.drop_index("ix_dobject_tmp_dtransform_id", table_name="dobject")
    op.drop_index("ix_dobject_tmp_file_suffix", table_name="dobject")
    op.drop_index("ix_dobject_tmp_name", table_name="dobject")
    op.drop_index("ix_dobject_tmp_storage_root", table_name="dobject")
    op.drop_index("ix_dobject_tmp_time_created", table_name="dobject")
    op.drop_index("ix_dobject_tmp_time_updated", table_name="dobject")
    op.create_index(op.f("ix_dobject_dtransform_id"), "dobject", ["dtransform_id"], unique=False)
    op.create_index(op.f("ix_dobject_file_suffix"), "dobject", ["file_suffix"], unique=False)
    op.create_index(op.f("ix_dobject_name"), "dobject", ["name"], unique=False)
    op.create_index(op.f("ix_dobject_storage_id"), "dobject", ["storage_id"], unique=False)
    op.create_index(op.f("ix_dobject_time_created"), "dobject", ["time_created"], unique=False)
    op.create_index(op.f("ix_dobject_time_updated"), "dobject", ["time_updated"], unique=False)
    op.add_column(
        "pipeline_run",
        sa.Column("pipeline_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    )
    op.add_column(
        "pipeline_run",
        sa.Column("pipeline_v", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
    )
    op.add_column(
        "pipeline_run",
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
    )
    op.create_index(op.f("ix_pipeline_run_name"), "pipeline_run", ["name"], unique=False)
    op.create_index(
        op.f("ix_pipeline_run_pipeline_id"),
        "pipeline_run",
        ["pipeline_id"],
        unique=False,
    )
    op.create_index(op.f("ix_pipeline_run_pipeline_v"), "pipeline_run", ["pipeline_v"], unique=False)
    with op.batch_alter_table("pipeline") as batch_op:
        batch_op.create_foreign_key("pipeline_run", "pipeline", ["pipeline_id", "pipeline_v"], ["id", "v"])
    op.add_column("storage", sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
    op.create_index(op.f("ix_storage_root"), "storage", ["root"], unique=False)
    op.alter_column("user", "handle", existing_type=sa.VARCHAR())
    with op.batch_alter_table("user") as batch_op:
        batch_op.create_unique_constraint("user", ["handle"])


def downgrade() -> None:
    pass



"""v0.28.1."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op

revision = "6952574e2d49"
down_revision = "1dafcf0b22aa"


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    if sqlite:
        with op.batch_alter_table(f"{prefix}dfolder", schema=schema) as batch_op:
            batch_op.add_column(sa.Column("_folderkey", sqm.sql.sqltypes.AutoString(), nullable=True))
            batch_op.create_index(batch_op.f("ix_core.dfolder__folderkey"), ["_folderkey"], unique=False)
    else:
        op.add_column(
            f"{prefix}dfolder",
            sa.Column("_folderkey", sqm.sql.sqltypes.AutoString(), nullable=True),
            schema=schema,
        )
        op.create_index(
            op.f("ix_core_dfolder__folderkey"),
            "dfolder",
            ["_folderkey"],
            unique=False,
            schema=schema,
        )


def downgrade() -> None:
    pass



import os
import re
from pathlib import Path, PurePath
from typing import Dict, List, Optional, Tuple, Union

import lnschema_core
import nbproject
from lamin_logger import logger
from lndb import settings
from lndb.dev import InstanceSettings
from lnschema_core import Run, Transform
from nbproject._is_run_from_ipython import is_run_from_ipython

from .dev.db._add import _private_not_empty

msg_init_complete = (
    "⚠️ Destructive operation! ⚠️\n\nAre you sure you saved the notebook before running"
    " `ln.track()`?\n - If not, hit save, and *overwrite* the notebook file.\n - If"
    " yes, hit save, and *discard* editor content.\n\nConsider using Jupyter Lab or"
    " Notebook for a seamless interactive notebook tracking experience."
)

msg_path_failed = (
    "Failed to infer notebook path.\nFix: Either track manually via"
    " `ln.track(ln.Transform(name='My notebook'))` or pass"
    " `notebook_path` to ln.track()."
)


def _write_notebook_meta(metadata):
    from nbproject._header import _env, _filepath

    nbproject.dev._frontend_commands._save_notebook(_env)
    nb = nbproject.dev.read_notebook(_filepath)
    nb.metadata["nbproject"] = metadata

    # write proper execution count
    header_re = re.compile(r"^[^#]*track\(", flags=re.MULTILINE)
    ccount = 0
    for cell in nb.cells:
        if cell["cell_type"] != "code":
            continue
        elif cell["execution_count"] is not None:
            ccount = cell["execution_count"]
        if header_re.match("".join(cell["source"])) is not None:
            cell["execution_count"] = ccount + 1
            break

    nbproject.dev.write_notebook(nb, _filepath)
    nbproject.dev._frontend_commands._reload_notebook(_env)


def reinitialize_notebook(
    id: str, name: str, metadata: Optional[Dict] = None
) -> Tuple[Transform, Dict]:
    from nbproject._header import _env, _filepath

    new_id, new_version = id, None
    if "NBPRJ_TEST_NBPATH" not in os.environ:
        response = input("Do you want to generate a new id? (y/n)")
    else:
        response = "y"
    if response == "y":
        new_id = lnschema_core.dev.id.transform()
    else:
        response = input(
            "Do you want to set a new version (e.g. '1.1')? Type 'n' for"
            " 'no'. (version/n)"
        )
        if response != "n":
            if new_version == "y":
                response = input("Please type the version: ")
            new_version = response

    nb = None
    if metadata is None:
        nb = nbproject.dev.read_notebook(_filepath)
        metadata = nb.metadata["nbproject"]

    metadata["id"] = new_id
    if new_version is None:
        new_version = "0"
    metadata["version"] = new_version

    # in "lab" & "notebook", we push the metadata write to the end of track execution
    # by returning metadata below
    if _env not in ("lab", "notebook", "test"):
        if nb is None:
            nb = nbproject.dev.read_notebook(_filepath)
        nb.metadata["nbproject"] = metadata
        nbproject.dev.write_notebook(nb, _filepath)
        raise SystemExit(msg_init_complete)

    transform = Transform(id=new_id, version=new_version, name=name, type="notebook")
    return transform, metadata


# from https://stackoverflow.com/questions/61901628
def get_notebook_name_colab() -> str:
    from socket import gethostbyname, gethostname  # type: ignore

    from requests import get  # type: ignore

    ip = gethostbyname(gethostname())  # 172.28.0.12
    name = get(f"http://{ip}:9000/api/sessions").json()[0]["name"]
    return name.rstrip(".ipynb")


class context:
    """Global run context."""

    instance: Optional[InstanceSettings] = None
    """Current instance."""
    transform: Optional[Transform] = None
    """Current transform."""
    run: Optional[Run] = None
    """Current run."""

    # exposed to user as ln.track()
    @classmethod
    def _track(
        cls,
        transform: Optional[Transform] = None,
        *,
        new_run: Optional[bool] = None,
        notebook_path: Optional[str] = None,
        pypackage: Union[str, List[str], None] = None,
        editor: Optional[str] = None,
    ) -> None:
        """Track `Transform` & `Run` records for a notebook or pipeline.

        Adds these records to the DB and exposes them as
        `ln.context.transform` and `ln.context.run`.

        Call without a `transform` record or without arguments
        when tracking a Jupyter notebook.

        If a Jupyter notebook has no associated metadata, attempts to write
        metadata to disk.

        Args:
            transform: Can be "pipeline" or "notebook".
            new_run: If True, loads latest run of transform.
            pypackage: One or more python packages to track.
            notebook_path: Filepath of notebook. Only needed if inference fails.
            editor: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
        """
        cls.instance = settings.instance
        logger.info(f"Instance: {cls.instance.identifier}")
        logger.info(f"User: {settings.user.handle}")
        import lamindb as ln

        if is_run_from_ipython and transform is None:
            cls._track_notebook(
                pypackage=pypackage,
                filepath=notebook_path,
                editor=editor,
            )
        elif transform is None:
            raise ValueError("Pass `transform` to .track()!")
        else:
            if transform.id is not None:  # id based look-up
                if transform.version is None:
                    transform_exists = (
                        ln.select(Transform, id=transform.id)
                        .order_by(Transform.created_at.desc())
                        .first()
                    )
                else:
                    transform_exists = ln.select(
                        Transform, id=transform.id, version=transform.version
                    ).first()
            else:  # name based lookup
                if transform.version is None:
                    transform_exists = (
                        ln.select(Transform, name=transform.name)
                        .order_by(Transform.created_at.desc())
                        .first()
                    )
                else:
                    transform_exists = ln.select(
                        Transform, name=transform.name, version=transform.version
                    ).first()
            if transform_exists is None:
                transform_exists = ln.add(transform)
                logger.success(f"Added: {transform}")
            else:
                logger.info(f"Loaded: {transform_exists}")
            cls.transform = transform_exists

        # for notebooks, default to loading latest runs
        if new_run is None:
            if cls.transform.type.value == "notebook":  # type: ignore
                new_run = False
            else:
                new_run = True

        # this here uses cls.transform and writes cls.run
        # should probably change that design
        Run(load_latest=not new_run)
        # so, this is a hack:
        if (
            _private_not_empty(cls.run, "_ln_identity_key")  # type: ignore
            and cls.run._ln_identity_key is not None  # type: ignore
        ):
            cls.run._sa_instance_state.key = cls.run._ln_identity_key  # type: ignore

        # only for newly intialized notebooks
        if hasattr(cls, "_notebook_meta"):
            _write_notebook_meta(cls._notebook_meta)  # type: ignore
            del cls._notebook_meta  # type: ignore

    @classmethod
    def _track_notebook(
        cls,
        *,
        filepath: Optional[str] = None,
        pypackage: Optional[Union[str, List[str]]] = None,
        editor: Optional[str] = None,
    ):
        """Infer Jupyter notebook metadata and create `Transform` record.

        Args:
            pypackage: One or more python packages to track.
            filepath: Filepath of notebook. Only needed if automatic inference fails.
            editor: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
        """
        cls.instance = settings.instance
        from nbproject.dev._jupyter_communicate import notebook_path

        metadata = None
        needs_init = False
        reference = None
        if filepath is None:
            path_env = None
            try:
                path_env = notebook_path(return_env=True)
            except Exception:
                raise RuntimeError(msg_path_failed)
            if path_env is None:
                raise RuntimeError(msg_path_failed)
            notebook_path, _env = path_env
        else:
            notebook_path = filepath
        if isinstance(notebook_path, (Path, PurePath)):
            notebook_path = notebook_path.as_posix()
        if notebook_path.endswith("Untitled.ipynb"):
            raise RuntimeError("Please rename your notebook before tracking it")
        if notebook_path.startswith("/filedId="):
            # This is Google Colab!
            # google colab fileID looks like this
            # /fileId=1KskciVXleoTeS_OGoJasXZJreDU9La_l
            # we'll take the first 12 characters
            colab_id = notebook_path.replace("/filedId=", "")
            id = colab_id[:12]
            reference = f"colab_id: {colab_id}"
            name = get_notebook_name_colab()
            _env = "colab"
        else:
            try:
                metadata, needs_init = nbproject.header(
                    pypackage=pypackage,
                    filepath=notebook_path if filepath is None else filepath,
                    env=_env if editor is None else editor,
                    metadata_only=True,
                )
                # this contains filepath if the header was run successfully
                from nbproject._header import _env, _filepath  # type: ignore
            except Exception:
                nbproject_failed_msg = (
                    "Auto-retrieval of notebook name & title failed.\nPlease paste"
                    " error at: https://github.com/laminlabs/nbproject/issues/new"
                    " \n\nFix: Run `ln.track(ln.Transform(name='My notebook'))`"
                )
                raise RuntimeError(nbproject_failed_msg)

        import lamindb as ln

        if needs_init:
            if _env in ("lab", "notebook"):
                cls._notebook_meta = metadata  # type: ignore
            else:
                nb = nbproject.dev.read_notebook(_filepath)
                nb.metadata["nbproject"] = metadata
                nbproject.dev.write_notebook(nb, _filepath)
                raise SystemExit(msg_init_complete)

        if _env in ("lab", "notebook"):
            # save the notebook in case that title was updated
            # but notebook not saved
            nbproject.dev._frontend_commands._save_notebook(_env)

        if metadata is not None:
            id = metadata["id"]
            version = metadata["version"]
            name = Path(_filepath).stem
            title = nbproject.meta.live.title
        else:
            version = "0"
            title = None

        transform = ln.select(Transform, id=id, version=version).one_or_none()
        if transform is None:
            transform = Transform(
                id=id,
                version=version,
                name=name,
                title=title,
                reference=reference,
                type="notebook",
            )
            transform = ln.add(transform)
            logger.success(f"Added: {transform}")
        else:
            logger.info(f"Loaded: {transform}")
            if transform.name != name or transform.title != title:
                response = input(
                    "Updated notebook name and/or title: Do you want to assign a new id"
                    " or version? (y/n)"
                )
                if response == "y":
                    transform, metadata = reinitialize_notebook(
                        transform.id, name, metadata
                    )
                    if _env in ("lab", "notebook"):
                        cls._notebook_meta = metadata  # type: ignore
                transform.name = name
                transform.title = title
                ln.add(transform)
                if response == "y":
                    logger.success(f"Added: {transform}")
                else:
                    logger.success(f"Updated: {transform}")

        cls.transform = transform



"""Schema tools & overview.

Guide: :doc:`/guide/schema`

You can access mounted schema modules with domain-specific entities via
available via `ln.schema.<module>.<entity>`.

However, we recommend to import schema modules, e.g., like `import
lnschema_bionty as bt`.

.. autosummary::
   :toctree: .

   graph
   view
   dev

"""
import importlib as _importlib

from lndb import settings as _settings
from lndb.dev._setup_schema import (
    check_schema_version_and_import as _check_schema_version_and_import,
)
from lnschema_core import Features, Project, Run, Storage
from lnschema_core import Transform as _Transform
from lnschema_core import User, dev

from .. import _INSTANCE_SETUP


def _import_schema():
    for name in _settings.instance.schema:
        module = _check_schema_version_and_import(name)
        globals()[module._name] = module


if _INSTANCE_SETUP:
    _import_schema()

from ._core import graph, list_tables, view

list_entities = list_tables  # backward compat

Pipeline = _Transform
Notebook = _Transform



import traceback
from typing import List, Optional, Union, overload  # noqa

import sqlmodel as sqm
from lndb import settings
from lndb_storage import delete_storage
from lnschema_core import File
from lnschema_core.dev._storage import storage_key_from_file
from lnschema_core.link import RunIn

from ._logger import colors, logger
from .dev.db._select import select


@overload
def delete(
    record: sqm.SQLModel,
    delete_data_from_storage: Optional[bool] = None,
) -> None:
    ...


@overload
def delete(
    records: List[sqm.SQLModel],
    delete_data_from_storage: Optional[bool] = None,
) -> None:  # type: ignore
    ...


@overload
def delete(
    entity: sqm.SQLModel,
    delete_data_from_storage: Optional[bool] = None,
    **fields,
) -> None:  # type: ignore
    ...


def delete(  # type: ignore
    record: Union[sqm.SQLModel, List[sqm.SQLModel]],
    delete_data_from_storage: Optional[bool] = None,
    **fields,
) -> None:
    """Delete metadata records & files.

    Guide: :doc:`/guide/add-delete`.

    Args:
        record: One or multiple records as instances of `SQLModel`.
        delete_data_from_storage: Whether to delete data from storage.

    Returns:
        `None`

    Examples:

        Delete by record:

        >>> experiment = ln.select(Experiment, id=experiment_id).one()
        >>> ln.delete(experiment)

        Delete by fields:

        >>> ln.delete(Experiment, id=experiment_id)
        >>> # the result of is equivalent to 1)

        Delete files (delete the metadata record and the file in storage)

        >>> file = ln.select(File, id=file_id).one()
        >>> # deleting the metadata record occurs automatically
        >>> # you will be asked whether to delete the file from storage
        >>> # or pass boolean values to `delete_data_from_storage`
        >>> ln.delete(file, delete_data_from_storage)

    """
    if isinstance(record, list):
        records = record
    elif isinstance(record, sqm.SQLModel):
        records = [record]
    else:
        model = record
        results = select(model, **fields).one_or_none()
        if results is None:
            return results
        else:
            records = [results]

    settings.instance._cloud_sqlite_locker.lock()
    session = settings.instance.session()
    for record in records:
        if isinstance(record, File):
            # delete run_ins related to the file that's to be deleted
            run_ins = session.exec(sqm.select(RunIn).where(RunIn.file_id == record.id))
            for run_in in run_ins:
                session.delete(run_in)
            try:
                session.commit()
            except Exception:
                logger.warning("Deleting run inputs failed!")
                traceback.print_exc()
        session.delete(record)
        try:
            session.commit()
            logger.success(
                f"Deleted {colors.yellow(f'row {record}')} in"
                f" {colors.blue(f'table {type(record).__name__}')}."
            )
        except Exception:
            traceback.print_exc()
        if isinstance(record, File):
            storage_key = storage_key_from_file(record)

            if delete_data_from_storage is None:
                # ask to confirm deleting data from storage
                delete_dialog = (
                    "Confirm Delete: Are you sure you want to delete"
                    f" object {storage_key} from storage? (y/n)"
                )
                decide = input(f"   {delete_dialog}")
            else:
                decide = "y" if delete_data_from_storage else "n"

            if decide not in ("y", "Y", "yes", "Yes", "YES"):
                continue
            try:
                delete_storage(storage_key)
                logger.success(
                    f"Deleted {colors.yellow(f'object {storage_key}')} from storage."
                )
            except Exception:
                traceback.print_exc()
    session.close()
    settings.instance._update_cloud_sqlite_file()
    settings.instance._cloud_sqlite_locker.unlock()



import base64
import hashlib
from pathlib import Path
from typing import Set


def to_b64_str(bstr: bytes):
    b64 = base64.urlsafe_b64encode(bstr).decode().strip("=")
    return b64


# a lot to read about this: lamin-notes/2022/hashing
def hash_set(s: Set[str]) -> str:
    bstr = ":".join(sorted(s)).encode("utf-8")
    # as we're truncating at 20 b64, we choose md5 over sha512
    return to_b64_str(hashlib.md5(bstr).digest())[:20]


def hash_file(path: Path) -> str:
    # based on https://stackoverflow.com/questions/3431825/generating-an-md5-hash-of-a-file  # noqa
    hash_md5 = hashlib.md5()
    with open(path, "rb") as file:
        for chunk in iter(lambda: file.read(4096), b""):
            hash_md5.update(chunk)
    return to_b64_str(hash_md5.digest())



from lnschema_core import dev


def biometa() -> str:
    """Biometa: 21 base62.

    Biometa consists of tuples of biosample, experiment, readout, featureset.

    There are about as many biometa rows as files.

    21 characters (62**21=4e+37 possibilities) outperform UUID (2*122=5e+36).
    """
    return dev.id.base62(n_char=21)


def biosample() -> str:
    """Bio sample: 20 base62."""
    return dev.id.base62(n_char=20)


def techsample() -> str:
    """Tech sample: 20 base62."""
    return dev.id.base62(n_char=20)


def readout() -> str:
    """Readout types: 5 base62.

    EFO ontology.
    """
    return dev.id.base62(n_char=5)


def experiment() -> str:
    """Experiments: 19 base62."""
    return dev.id.base62(n_char=19)


def experiment_type() -> str:
    """Experiments: 12 base62."""
    return dev.id.base62(n_char=12)


def treatment() -> str:
    """Projects: 12 base62."""
    return dev.id.base62(n_char=12)



import warnings
from typing import Optional

import scipy.sparse as sparse
import zarr
from anndata import AnnData
from anndata._io import read_zarr
from anndata._io.specs import write_elem
from lndb.dev.upath import infer_filesystem

from .object._anndata_sizes import _size_elem, _size_raw, size_adata


def read_adata_zarr(storepath) -> AnnData:
    fs, storepath = infer_filesystem(storepath)

    store = fs.get_mapper(storepath, check=True)
    adata = read_zarr(store)

    return adata


def write_adata_zarr(
    adata: AnnData, storepath, callback=None, chunks=None, **dataset_kwargs
):
    fs, storepath = infer_filesystem(storepath)

    store = fs.get_mapper(storepath, create=True)
    f = zarr.open(store, mode="w")

    adata.strings_to_categoricals()
    if adata.raw is not None:
        adata.strings_to_categoricals(adata.raw.var)

    f.attrs.setdefault("encoding-type", "anndata")
    f.attrs.setdefault("encoding-version", "0.1.0")

    adata_size = None
    cumulative_val = 0

    def _cb(key_write: Optional[str] = None):
        nonlocal adata_size
        nonlocal cumulative_val

        if callback is None:
            return None
        if adata_size is None:
            adata_size = size_adata(adata)
        if key_write is None:
            # begin or finish
            if cumulative_val < adata_size:
                callback(adata_size, adata_size if cumulative_val > 0 else 0)
            return None

        elem = getattr(adata, key_write, None)
        if elem is None:
            return None
        elem_size = _size_raw(elem) if key_write == "raw" else _size_elem(elem)
        if elem_size == 0:
            return None

        cumulative_val += elem_size
        callback(adata_size, cumulative_val)

    def _write_elem_cb(f, k, elem, dataset_kwargs):
        write_elem(f, k, elem, dataset_kwargs=dataset_kwargs)
        _cb(k)

    _cb(None)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, module="zarr")

        if chunks is not None and not isinstance(adata.X, sparse.spmatrix):
            _write_elem_cb(
                f, "X", adata.X, dataset_kwargs=dict(chunks=chunks, **dataset_kwargs)
            )
        else:
            _write_elem_cb(f, "X", adata.X, dataset_kwargs=dataset_kwargs)
        for elem in ("obs", "var"):
            _write_elem_cb(f, elem, getattr(adata, elem), dataset_kwargs=dataset_kwargs)
        for elem in ("obsm", "varm", "obsp", "varp", "layers", "uns"):
            _write_elem_cb(
                f, elem, dict(getattr(adata, elem)), dataset_kwargs=dataset_kwargs
            )
        _write_elem_cb(f, "raw", adata.raw, dataset_kwargs=dataset_kwargs)
    # todo: fix size less than total at the end
    _cb(None)



"""v0.3.0.

Revision ID: 8c78543d1c5b
Revises: 0560ee3d73dc
Create Date: 2022-08-08 15:45:42.737438

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "8c78543d1c5b"
down_revision = "0560ee3d73dc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Completely new table
    op.create_table(
        "pipeline",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.PrimaryKeyConstraint("id", "v"),
    )

    # New table where we can insert one row from the old schema_version
    op.create_table(
        "version_yvzi",
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("user_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("time_created", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("v"),
    )
    op.execute("insert into version_yvzi select id, user_id, time_created from schema_version")
    op.drop_table("schema_version")

    # New table where we need to insert a link to dobject and to jupynb
    op.create_table(
        "dtransform",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("jupynb_id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("jupynb_v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("pipeline_id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("pipeline_v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.ForeignKeyConstraint(
            ["jupynb_id", "jupynb_v"],
            ["jupynb.id", "jupynb.v"],
            name="dtransform_jupynb",
        ),
        sa.ForeignKeyConstraint(
            ["pipeline_id", "pipeline_v"],
            ["pipeline.id", "pipeline.v"],
            name="dtransform_pipeline",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "dtransform_in",
        sa.Column("dtransform_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dobject_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dobject_v", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dobject_id", "dobject_v"],
            ["dobject.id", "dobject.v"],
            name="dtransform_in_dobject",
        ),
        sa.ForeignKeyConstraint(
            ["dtransform_id"],
            ["dtransform.id"],
        ),
        sa.PrimaryKeyConstraint("dtransform_id", "dobject_id", "dobject_v"),
    )
    op.create_table(
        "dtransform_out",
        sa.Column("dtransform_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dobject_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dobject_v", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dobject_id", "dobject_v"],
            ["dobject.id", "dobject.v"],
            name="dtransform_out_dobject",
        ),
        sa.ForeignKeyConstraint(
            ["dtransform_id"],
            ["dtransform.id"],
        ),
        sa.PrimaryKeyConstraint("dtransform_id", "dobject_id", "dobject_v"),
    )
    op.execute("insert into dtransform (id, jupynb_id, jupynb_v) select id, id, v from jupynb")

    # Add the dsouce_id column and insert data
    op.create_table(
        "dobject_",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("v", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("file_suffix", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dsource_id", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("jupynb_id", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("jupynb_v", sqlmodel.sql.sqltypes.AutoString()),
        sa.Column("time_created", sa.DateTime(), nullable=False),
        sa.Column("time_updated", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dsource_id"],
            ["dtransform.id"],
        ),
        sa.PrimaryKeyConstraint("id", "v"),
    )
    op.execute(
        "insert into dobject_(id, v, name, file_suffix, jupynb_id, jupynb_v,"
        " time_created, time_updated) select id, v, name, file_suffix, jupynb_id,"
        " jupynb_v, time_created, time_updated from dobject"
    )
    op.drop_table("dobject")
    op.rename_table("dobject_", "dobject")
    op.execute("update dobject set dsource_id = (select id from dtransform where (dtransform.jupynb_id, dtransform.jupynb_v) = (dobject.jupynb_id, dobject.jupynb_v))")
    # now we don't need these two columns anymore on dobject
    op.drop_column("dobject", "jupynb_v")
    op.drop_column("dobject", "jupynb_id")

    # Populate dtransform_out
    op.execute("insert into dtransform_out (dtransform_id, dobject_id, dobject_v) select dsource_id, id, v from dobject")

    # This is a rename from track_do, so we got to insert all data from track_do
    op.create_table(
        "usage",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("user_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("time", sa.DateTime(), nullable=False),
        sa.Column("dobject_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("dobject_v", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dobject_id", "dobject_v"],
            ["dobject.id", "dobject.v"],
            name="usage_dobject",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.execute("insert into usage (id, type, user_id, time, dobject_id, dobject_v) select id, type, user_id, time, dobject_id, dobject_v from track_do")
    op.drop_table("track_do")

    # jupynb table no longer has type
    op.drop_column("jupynb", "type")

    # User table gets a handle now
    op.add_column("user", sa.Column("handle", sqlmodel.sql.sqltypes.AutoString()))
    op.execute("update user set handle = 'falexwolf' where id = 'FBa7SHjn'")


def downgrade() -> None:
    pass



"""Small example datasets.

.. autosummary::
   :toctree: .

   file_fcs
   file_jpg_paradisi05
   file_fastq
   file_bam
   file_mini_csv
   dir_scrnaseq_cellranger
   generate_cell_ranger_files
   anndata_mouse_sc_lymph_node
   anndata_human_immune_cells
   anndata_pbmc68k_reduced
   anndata_pbmc3k_processed
   anndata_with_obs
"""

from ._core import (
    anndata_human_immune_cells,
    anndata_mouse_sc_lymph_node,
    anndata_pbmc3k_processed,
    anndata_pbmc68k_reduced,
    anndata_with_obs,
    dir_scrnaseq_cellranger,
    file_bam,
    file_fastq,
    file_fcs,
    file_jpg_paradisi05,
    file_mini_csv,
    generate_cell_ranger_files,
)



"""v0.25.9.

Revision ID: 8bf788467d0a
Revises: 9d283a1685a5
Create Date: 2023-02-07 14:56:26.461590

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8bf788467d0a"
down_revision = "9d283a1685a5"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    op.rename_table(old_table_name=f"{prefix}dset", new_table_name=f"{prefix}dfolder", schema=schema)
    op.rename_table(
        old_table_name=f"{prefix}dset_dobject",
        new_table_name=f"{prefix}dfolder_dobject",
        schema=schema,
    )
    op.rename_table(
        old_table_name=f"{prefix}project_dset",
        new_table_name=f"{prefix}project_dfolder",
        schema=schema,
    )
    op.alter_column(
        f"{prefix}dfolder_dobject",
        column_name="dset_id",
        new_column_name="dfolder_id",
        schema=schema,
    )
    op.alter_column(
        f"{prefix}project_dfolder",
        column_name="dset_id",
        new_column_name="dfolder_id",
        schema=schema,
    )
    if sqlite:
        op.drop_index("ix_core.dset_created_at", table_name=f"{prefix}dfolder", schema=schema)
        op.drop_index("ix_core.dset_created_by", table_name=f"{prefix}dfolder", schema=schema)
        op.drop_index("ix_core.dset_name", table_name=f"{prefix}dfolder", schema=schema)
        op.drop_index("ix_core.dset_updated_at", table_name=f"{prefix}dfolder", schema=schema)
        op.create_index(
            op.f("ix_core.dfolder_created_at"),
            f"{prefix}dfolder",
            ["created_at"],
            unique=False,
            schema=schema,
        )
        op.create_index(
            op.f("ix_core.dfolder_created_by"),
            f"{prefix}dfolder",
            ["created_by"],
            unique=False,
            schema=schema,
        )
        op.create_index(
            op.f("ix_core.dfolder_name"),
            f"{prefix}dfolder",
            ["name"],
            unique=False,
            schema=schema,
        )
        op.create_index(
            op.f("ix_core.dfolder_updated_at"),
            f"{prefix}dfolder",
            ["updated_at"],
            unique=False,
            schema=schema,
        )
    else:
        op.drop_index("ix_core_dset_created_at", table_name=f"{prefix}dfolder", schema=schema)
        op.drop_index("ix_core_dset_created_by", table_name=f"{prefix}dfolder", schema=schema)
        op.drop_index("ix_core_dset_name", table_name=f"{prefix}dfolder", schema=schema)
        op.drop_index("ix_core_dset_updated_at", table_name=f"{prefix}dfolder", schema=schema)
        op.create_index(
            op.f("ix_core_dfolder_created_at"),
            f"{prefix}dfolder",
            ["created_at"],
            unique=False,
            schema=schema,
        )
        op.create_index(
            op.f("ix_core_dfolder_created_by"),
            f"{prefix}dfolder",
            ["created_by"],
            unique=False,
            schema=schema,
        )
        op.create_index(
            op.f("ix_core_dfolder_name"),
            f"{prefix}dfolder",
            ["name"],
            unique=False,
            schema=schema,
        )
        op.create_index(
            op.f("ix_core_dfolder_updated_at"),
            f"{prefix}dfolder",
            ["updated_at"],
            unique=False,
            schema=schema,
        )


def downgrade() -> None:
    pass



from pathlib import Path

from upath import UPath

from lamindb._file import get_check_path_in_storage, get_relative_path_to_root


def test_get_name_suffix_from_filepath():
    # based on https://stackoverflow.com/questions/31890341/clean-way-to-get-the-true-stem-of-a-path-object  # noqa
    dataset = [
        ("a", "a", ""),
        ("a.txt", "a", ".txt"),
        ("archive.tar.gz", "archive", ".tar.gz"),
        ("directory/file", "file", ""),
        ("d.x.y.z/f.a.b.c", "f", ".a.b.c"),
        ("logs/date.log.txt", "date", ".log.txt"),
    ]
    for path, _, suffix in dataset:
        filepath = Path(path)
        assert suffix == "".join(filepath.suffixes)


def test_storage_root_upath_equivalence():
    storage_root = UPath("s3://lamindb-ci")
    filepath = UPath("s3://lamindb-ci/test-data/Species.csv")
    assert filepath.parents[-1] == storage_root


def test_get_relative_path_to_root():
    # upath on S3
    root = UPath("s3://lamindb-ci")
    upath = UPath("s3://lamindb-ci/test-data/test.csv")
    assert (
        "test-data/test.csv" == get_relative_path_to_root(upath, root=root).as_posix()
    )
    # local path
    root = Path("/lamindb-ci")
    upath = Path("/lamindb-ci/test-data/test.csv")
    assert (
        "test-data/test.csv" == get_relative_path_to_root(upath, root=root).as_posix()
    )


def test_get_check_path_in_storage():
    # UPath
    root = UPath("s3://lamindb-ci")
    upath = UPath("s3://lamindb-ci/test-data/test.csv")
    assert get_check_path_in_storage(upath, root=root)
    upath2 = UPath("s3://lamindb-setup/test-data/test.csv")
    assert not get_check_path_in_storage(upath2, root=root)
    # local path
    root = Path("/lamindb-ci")
    path = Path("/lamindb-ci/test-data/test.csv")
    assert get_check_path_in_storage(path, root=root)
    path = Path("/lamindb-other/test-data/test.csv")
    assert not get_check_path_in_storage(path, root=root)
    # Local & UPath
    root = UPath("s3://lamindb-ci")
    path = Path("/lamindb-ci/test-data/test.csv")
    assert not get_check_path_in_storage(path, root=root)



from pathlib import Path

from lndb._migrate import generate_module_files
from lndb.test import (
    get_package_name,
    migrate_clones,
    migration_id_is_consistent,
    model_definitions_match_ddl,
)

from lnschema_core import _schema_id as schema_id

package_name = get_package_name()
migrations_path = Path(__file__).parent.parent / package_name / "migrations"


def test_migration_id_is_consistent():
    assert migration_id_is_consistent(package_name)


def test_model_definitions_match_ddl_postgres():
    generate_module_files(package_name=package_name, migrations_path=migrations_path, schema_id=schema_id)
    model_definitions_match_ddl(package_name, dialect_name="postgresql")


def test_migrate_clones_sqlite():
    results = migrate_clones(package_name, n_instances=1, dialect_name="sqlite")
    if "migrate-failed" in results:
        raise RuntimeError("Migration e2e test failed.")


def test_migrate_clones_postgres():
    results = migrate_clones(package_name, n_instances=1, dialect_name="postgresql")
    if "migrate-failed" in results:
        raise RuntimeError("Migration e2e test failed.")



from datetime import datetime as datetime
from typing import Optional

from lnschema_core._timestamps import CreatedAt, UpdatedAt
from lnschema_core._users import CreatedBy
from lnschema_core.dev.sqlmodel import schema_sqlmodel
from sqlmodel import Field

from . import _name as schema_name
from .dev import id as idg

SQLModel, prefix, schema_arg = schema_sqlmodel(schema_name)


class ExperimentBase(SQLModel):  # type: ignore
    """Experiments."""

    id: str = Field(default_factory=idg.experiment, primary_key=True)
    name: str = Field(default=None, index=True)
    date: datetime = Field(default=None, index=True)
    """Date on which the experiment is performed."""
    created_by: str = CreatedBy
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class ExperimentTypeBase(SQLModel):  # type: ignore
    """Experiment types.

    Useful for bucketing experiments.
    """

    id: str = Field(default_factory=idg.experiment_type, primary_key=True)
    name: str = Field(default=None, index=True)
    efo_id: str = Field(default=None, unique=True)
    created_by: str = CreatedBy
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class BiosampleBase(SQLModel):  # type: ignore
    """Biosamples: the fundamental observational unit."""

    id: str = Field(default_factory=idg.biosample, primary_key=True)
    name: Optional[str] = Field(default=None, index=True)
    created_by: str = CreatedBy
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class TechsampleBase(SQLModel):  # type: ignore
    """Tech samples that are generated due to instrument units."""

    id: str = Field(default_factory=idg.techsample, primary_key=True)
    name: Optional[str] = Field(default=None, index=True)
    batch: Optional[str] = None
    filepath_r1: Optional[str] = None
    filepath_r2: Optional[str] = None
    created_by: str = CreatedBy
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt



"""v0.8.1.

Revision ID: 1190648443cb
Revises: 7e8f7b30792e
Create Date: 2022-09-26 13:39:13.069980
"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1190648443cb"
down_revision = "7e8f7b30792e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("jupynb", schema=None) as batch_op:
        batch_op.alter_column("name", existing_type=sa.VARCHAR(), nullable=False)


def downgrade() -> None:
    with op.batch_alter_table("jupynb", schema=None) as batch_op:
        batch_op.alter_column("name", existing_type=sa.VARCHAR(), nullable=True)



"""v0.5.1.

Revision ID: d1b3e5da6391
Revises: 3badf20f18c8
Create Date: 2022-08-29 19:35:13.905660

"""
import sqlalchemy as sa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d1b3e5da6391"
down_revision = "3badf20f18c8"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("dobject") as batch_op:
        batch_op.drop_column("storage_root")
    op.add_column("pipeline", sa.Column("time_created", sa.DateTime(), nullable=False))
    op.add_column("pipeline_run", sa.Column("time_created", sa.DateTime(), nullable=False))
    with op.batch_alter_table("pipeline") as batch_op:
        batch_op.create_foreign_key("pipeline_run", "pipeline", ["pipeline_id", "pipeline_v"], ["id", "v"])


def downgrade() -> None:
    pass



"""IDs.

Entity-related generators:

.. autosummary::
   :toctree: .

   biometa
   techsample
   biosample
   readout
   treatment
   experiment
   experiment_type

"""
from ._id import (  # noqa
    biometa,
    biosample,
    experiment,
    experiment_type,
    readout,
    techsample,
    treatment,
)



from datetime import datetime as datetime
from typing import Optional

from lnschema_core._timestamps import CreatedAt
from lnschema_core._users import CreatedBy
from sqlmodel import Field, SQLModel


class version_vvhc(SQLModel, table=True):  # type: ignore
    """Wetlab schema module versions deployed in a given instance.

    Migrations of the schema module add rows to this table, storing the schema
    module version to which we migrated along with the user who performed the
    migration.
    """

    v: str = Field(primary_key=True)
    migration: Optional[str] = None
    user_id: str = CreatedBy
    created_at: datetime = CreatedAt


class migration_vvhc(SQLModel, table=True):  # type: ignore
    """Latest migration.

    This stores the reference to the latest migration script deployed.
    """

    version_num: Optional[str] = Field(primary_key=True)



from pathlib import Path
from subprocess import run

import lndb
import pytest
from lamin_logger import logger


def pytest_sessionstart(session: pytest.Session):
    instance_dirs = [
        d for d in ["./docs/guide/mydata", "./mydata-test-db"] if Path(d).exists()
    ]
    for instance_dir in instance_dirs:
        cmd = f"rm -r {instance_dir}"
        try:
            run(cmd)
            logger.info(cmd)
        except Exception:
            logger.info(f"Could not delete {instance_dir}")
    cmd = "lamin init --storage mydata-test-db"
    logger.info(cmd)
    run(cmd, shell=True)
    try:
        lndb.load("testuser1/lamindb-ci", migrate=True)
        lndb.delete("lamindb-ci")
    except Exception:
        logger.info("Could not delete testuser1/lamindb-ci")



"""v0.16.0.

Revision ID: 4ee426b656bb
Revises: 98da12fc80a8
Create Date: 2022-11-10 15:09:46.416764

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4ee426b656bb"
down_revision = "98da12fc80a8"
branch_labels = None
depends_on = None


def upgrade_sqlite() -> None:
    op.rename_table("core.pipeline_run", "core.run")
    op.alter_column("core.dtransform", column_name="pipeline_run_id", new_column_name="run_id")

    # clean up names
    with op.batch_alter_table("core.dobject", schema=None) as batch_op:
        batch_op.drop_index("ix_dobject_checksum")
        batch_op.drop_index("ix_dobject_created_at")
        batch_op.drop_index("ix_dobject_dtransform_id")
        batch_op.drop_index("ix_dobject_name")
        batch_op.drop_index("ix_dobject_size")
        batch_op.drop_index("ix_dobject_storage_id")
        batch_op.drop_index("ix_dobject_suffix")
        batch_op.drop_index("ix_dobject_updated_at")
        batch_op.create_index(batch_op.f("ix_core.dobject_checksum"), ["checksum"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dobject_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dobject_dtransform_id"), ["dtransform_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dobject_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dobject_size"), ["size"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dobject_storage_id"), ["storage_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dobject_suffix"), ["suffix"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dobject_updated_at"), ["updated_at"], unique=False)

    with op.batch_alter_table("core.dset", schema=None) as batch_op:
        batch_op.drop_index("ix_dset_created_at")
        batch_op.drop_index("ix_dset_created_by")
        batch_op.drop_index("ix_dset_name")
        batch_op.drop_index("ix_dset_updated_at")
        batch_op.create_index(batch_op.f("ix_core.dset_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dset_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dset_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dset_updated_at"), ["updated_at"], unique=False)

    with op.batch_alter_table("core.dtransform", schema=None) as batch_op:
        batch_op.drop_index("ix_dtransform_jupynb_id")
        batch_op.drop_index("ix_dtransform_jupynb_v")
        batch_op.drop_index("ix_dtransform_pipeline_run_id")
        batch_op.create_index(batch_op.f("ix_core.dtransform_jupynb_id"), ["jupynb_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dtransform_jupynb_v"), ["jupynb_v"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.dtransform_run_id"), ["run_id"], unique=False)
        # the following line should be built in already into the alter_column in l. 21
        # batch_op.create_foreign_key(
        #    batch_op.f('fk_core.dtransform_run_id_run'), 'core.run', ['run_id'], ['id']
        # )

    with op.batch_alter_table("core.jupynb", schema=None) as batch_op:
        batch_op.drop_index("ix_jupynb_created_at")
        batch_op.drop_index("ix_jupynb_created_by")
        batch_op.drop_index("ix_jupynb_name")
        batch_op.drop_index("ix_jupynb_updated_at")
        batch_op.create_index(batch_op.f("ix_core.jupynb_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.jupynb_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.jupynb_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.jupynb_updated_at"), ["updated_at"], unique=False)

    with op.batch_alter_table("core.pipeline", schema=None) as batch_op:
        batch_op.drop_index("ix_pipeline_created_at")
        batch_op.drop_index("ix_pipeline_created_by")
        batch_op.drop_index("ix_pipeline_name")
        batch_op.drop_index("ix_pipeline_reference")
        batch_op.drop_index("ix_pipeline_updated_at")
        batch_op.create_index(batch_op.f("ix_core.pipeline_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.pipeline_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.pipeline_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.pipeline_reference"), ["reference"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.pipeline_updated_at"), ["updated_at"], unique=False)

    with op.batch_alter_table("core.project", schema=None) as batch_op:
        batch_op.drop_index("ix_project_created_at")
        batch_op.drop_index("ix_project_created_by")
        batch_op.drop_index("ix_project_name")
        batch_op.drop_index("ix_project_updated_at")
        batch_op.create_index(batch_op.f("ix_core.project_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.project_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.project_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.project_updated_at"), ["updated_at"], unique=False)

    with op.batch_alter_table("core.run", schema=None) as batch_op:
        batch_op.drop_index("ix_pipeline_run_created_at")
        batch_op.drop_index("ix_pipeline_run_created_by")
        batch_op.drop_index("ix_pipeline_run_name")
        batch_op.drop_index("ix_pipeline_run_pipeline_id")
        batch_op.drop_index("ix_pipeline_run_pipeline_v")
        batch_op.create_index(batch_op.f("ix_core.run_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_pipeline_id"), ["pipeline_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_pipeline_v"), ["pipeline_v"], unique=False)

    with op.batch_alter_table("core.storage", schema=None) as batch_op:
        batch_op.drop_index("ix_storage_created_at")
        batch_op.drop_index("ix_storage_root")
        batch_op.drop_index("ix_storage_updated_at")
        batch_op.create_index(batch_op.f("ix_core.storage_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.storage_root"), ["root"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.storage_updated_at"), ["updated_at"], unique=False)

    with op.batch_alter_table("core.usage", schema=None) as batch_op:
        batch_op.drop_index("ix_usage_dobject_id")
        batch_op.drop_index("ix_usage_time")
        batch_op.drop_index("ix_usage_type")
        batch_op.drop_index("ix_usage_user_id")
        batch_op.create_index(batch_op.f("ix_core.usage_dobject_id"), ["dobject_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.usage_time"), ["time"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.usage_type"), ["type"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.usage_user_id"), ["user_id"], unique=False)

    with op.batch_alter_table("core.user", schema=None) as batch_op:
        batch_op.drop_index("ix_user_created_at")
        batch_op.drop_index("ix_user_email")
        batch_op.drop_index("ix_user_handle")
        batch_op.drop_index("ix_user_name")
        batch_op.drop_index("ix_user_updated_at")
        batch_op.create_index(batch_op.f("ix_core.user_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.user_email"), ["email"], unique=True)
        batch_op.create_index(batch_op.f("ix_core.user_handle"), ["handle"], unique=True)
        batch_op.create_index(batch_op.f("ix_core.user_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.user_updated_at"), ["updated_at"], unique=False)


def upgrade_postgres():
    op.rename_table("pipeline_run", "run", schema="core")
    op.alter_column(
        "dtransform",
        column_name="pipeline_run_id",
        new_column_name="run_id",
        schema="core",
    )


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        upgrade_sqlite()
    else:
        upgrade_postgres()


def downgrade() -> None:
    pass



from typing import List, Optional, Union

from anndata import AnnData, concat
from lamin_logger import logger
from lnschema_core import File

from .object import LazySelector, _subset_anndata_file

SUFFIXES = (".h5ad", ".zarr")


def subset(
    files: Union[List[File], File],
    query_obs: Optional[Union[List[str], str, LazySelector]] = None,
    query_var: Optional[Union[List[str], str, LazySelector]] = None,
    use_concat: bool = False,
    concat_args: Optional[dict] = None,
) -> Union[List[AnnData], AnnData, None]:
    """Subset AnnData files and stream results into memory.

    Args:
        files: A `File` or a list of `Files` containing `AnnData` objects
        to subset and load into memory.
        query_obs: The pandas query string to evaluate on `.obs` of each
        underlying `AnnData` object.
        query_var: The pandas query string to evaluate on `.var` of each
        underlying `AnnData` object.
        use_concat: If `True`, applies `anndata.concat` on
        the returned `AnnData` objects.
        concat_args: Arguments for concatenation.
    """
    if isinstance(files, File):
        files = [files]

    n_files = len(files)
    if isinstance(query_obs, list):
        if len(query_obs) != n_files:
            raise ValueError("query_obs list should be the same length as files.")
    else:
        query_obs = [query_obs] * n_files  # type: ignore
    if isinstance(query_var, list):
        if len(query_var) != n_files:
            raise ValueError("query_var list should be the same length as files.")
    else:
        query_var = [query_var] * n_files  # type: ignore

    adatas = []
    # todo: implement parallel processing
    for i, file in enumerate(files):
        if file.suffix not in SUFFIXES:
            logger.warning(f"File {file.id} is not an AnnData object, ignoring.")
            continue
        result = _subset_anndata_file(file, query_obs[i], query_var[i])
        if result is not None:
            adatas.append(result)

    if not use_concat:
        return adatas
    # concat branch here
    n_adatas = len(adatas)
    if n_adatas == 0:
        return None
    elif n_adatas == 1:
        return adatas[0]
    else:
        concat_args = {} if concat_args is None else concat_args
        return concat(adatas, **concat_args)



import anndata
from anndata import AnnData
from lndb import settings
from lndb.dev.upath import infer_filesystem


def h5ad_to_anndata(filekey) -> AnnData:
    """h5ad → AnnData."""
    return anndata.read(settings.instance.storage.local_filepath(filekey))


def read_adata_h5ad(filepath, **kwargs) -> AnnData:
    fs, filepath = infer_filesystem(filepath)

    with fs.open(filepath, mode="rb") as file:
        adata = anndata.read_h5ad(file, backed=False, **kwargs)
        return adata



"""Helpers for the db API the database.

Select:

.. autosummary::
   :toctree: .

   SelectStmt
   ExecStmt
"""

from ._add import add
from ._select import ExecStmt, SelectStmt, select
from ._session import Session



from typing import Callable, List, TypeVar

import numpy as np
import pandas as pd
from lamin_logger import logger
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlmodel import SQLModel

from .dev.db._select import select

ListLike = TypeVar("ListLike", pd.Series, list, np.array)
SQLModelField = TypeVar("SQLModelField", Callable, InstrumentedAttribute)


def parse(
    iterable: ListLike, field: SQLModelField, from_bionty: bool = True
) -> List[SQLModel]:
    """Parse a dataset column based on a SQLModel entity field.

    Guide: :doc:`/guide/parse`.

    Args:
        iterable: a `ListLike` of values.
        field: a `SQLModel` field that correspond to the input list.
        from_bionty: whether to use bionty for auto-completion of the fields.

    Returns:
        A list of SQLModel records.
    """
    records = []
    entity = field.class_  # type:ignore
    for i in set(iterable):
        # No entries are made for NAs, '', None
        if pd.isnull(i) or i == "":
            continue
        kwargs = {field.name: i}  # type:ignore
        record = _create_record(entity, **kwargs, from_bionty=from_bionty)
        records.append(record)
    return records


def _create_record(entity: SQLModel, from_bionty: bool = True, **kwargs):
    """Create a record from bionty with configured ontologies."""
    record = select(entity, **kwargs).one_or_none()
    if record is None:
        if from_bionty:
            try:
                record = entity.from_bionty(**kwargs)
            except AttributeError:
                record = entity(**kwargs)
            except (ValueError, KeyError):
                logger.warning(
                    f"No entry found in bionty with {kwargs}!"
                    " Couldn't populate additional fields..."
                )
                record = entity(**kwargs)
        else:
            record = entity(**kwargs)
    return record



"""v0.21.3.

Revision ID: 9d94f3b9566d
Revises: 4b4005b7841c
Create Date: 2022-12-07 00:12:51.485117

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9d94f3b9566d"
down_revision = "4b4005b7841c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    with op.batch_alter_table(f"{prefix}dobject", schema=schema) as batch_op:
        batch_op.alter_column(
            "size",
            existing_type=sa.FLOAT(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )


def downgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    with op.batch_alter_table(f"{prefix}dobject", schema=schema) as batch_op:
        batch_op.alter_column(
            "size",
            existing_type=sa.BigInteger(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )



import os
from pathlib import Path

from anndata import AnnData
from lamin_logger import logger
from lndb import settings
from typeguard import typechecked


@typechecked
def anndata_to_h5ad(adata: AnnData, filekey: str) -> Path:
    """AnnData → h5ad."""
    path = settings.instance.storage.key_to_filepath(filekey)
    if settings.instance.storage.is_cloud:
        cache_file = settings.instance.storage.cloud_to_local_no_update(path)  # type: ignore  # noqa
        cache_file.parent.mkdir(exist_ok=True)
        logger.debug(f"Writing cache file: {cache_file}.")
        adata.write(cache_file)
        logger.debug("Uploading cache file.")
        path.upload_from(cache_file)  # type: ignore
        # to avoid download from the cloud within synchronization
        mtime = path.modified.timestamp()
        os.utime(cache_file, times=(mtime, mtime))
    else:
        adata.write(path)
        cache_file = path
    return cache_file



"""Developer API.

.. autosummary::
   :toctree: .

   db
   LazyDataFrame
   datasets

Utilities:

.. autosummary::
   :toctree: .

   doc_args
"""

from lndb_storage.object._lazy_field import Lazy as LazyDataFrame
from lnschema_core.dev import id  # noqa

from . import datasets  # noqa
from . import db  # noqa
from ._docs import doc_args  # noqa



"""SQLModel children.

.. autosummary::
   :toctree: .

   SQLModelModule
   SQLModelPrefix
"""

import importlib
import re
import typing
from collections import namedtuple
from typing import Any, Iterable, Optional, Sequence, Tuple, Union

import lndb
import sqlmodel as sqm
from pydantic import create_model
from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import relationship as sa_relationship
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.orm.session import object_session
from typeguard import check_type

# add naming convention for alembic
sqm.SQLModel.metadata.naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_`%(constraint_name)s`",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


# it's tricky to deal with class variables in SQLModel children
# hence, we're using a global variable, here
SCHEMA_NAME = None


def __repr_args__(self) -> Sequence[Tuple[Optional[str], Any]]:
    # sort like fields
    return [(k, self.__dict__[k]) for k in self.__fields__ if (not k.startswith("_sa_") and k in self.__dict__ and self.__dict__[k] is not None)]  # noqa  # noqa


def __repr__(self):
    if object_session(self) is not None:
        return "[session open] " + super(sqm.SQLModel, self).__repr__()
    else:
        return super(sqm.SQLModel, self).__repr__()


sqm.SQLModel.__repr_args__ = __repr_args__
sqm.SQLModel.__repr__ = __repr__
sqm.SQLModel.__str__ = __repr__


class SQLModelLamin(sqm.SQLModel):  # type: ignore
    """SQLModel at Lamin."""

    def __init__(self, **user_kwargs):
        # pydantic does not enforce strict type checking for complex types, only their attributes
        validate_relationship_types(self, user_kwargs)
        super().__init__(**user_kwargs)
        validate_with_pydantic(self, user_kwargs)

    @classmethod
    def df(cls):
        import lamindb as ln

        return ln.select(cls).df()

    @classmethod
    def lookup(cls, field: Optional[str] = None):
        if field is None:
            # by default use the name field
            if "name" in cls.__fields__:
                field = "name"
            else:
                non_ids = [i for i in cls.__fields__.keys() if "id" not in i]
                if len(non_ids) > 0:
                    # the first field isn't named with id
                    field = non_ids[0]
                else:
                    # the first field
                    field = next(iter(cls.__fields__.keys()))
        values = set(cls.df()[field].values)
        keys = _to_lookup_keys(values, padding=cls.__name__)
        return _namedtuple_from_dict(d=dict(zip(keys, values)), name=cls.__name__)


class SQLModelSchemaModule(SQLModelLamin):
    """SQLModel for schema module."""

    def __init__(self, **user_kwargs):
        super().__init__(**user_kwargs)

    # this here is problematic for those tables that overwrite
    # __table_args__; we currently need to treat them manually
    @declared_attr
    def __table_args__(cls) -> str:
        """Update table args with schema module."""
        return dict(schema=f"{SCHEMA_NAME}")  # type: ignore


class SQLModelSchemaModulePrefix(SQLModelLamin):
    """SQLModel for schema module with prefixed table names."""

    def __init__(self, **user_kwargs):
        super().__init__(**user_kwargs)

    @declared_attr
    def __tablename__(cls) -> str:  # type: ignore
        """Prefix table name with schema module."""
        return f"{SCHEMA_NAME}.{cls.__name__.lower()}"


def is_sqlite():
    return lndb.settings.instance.dialect == "sqlite"


def schema_sqlmodel(schema_name: str):
    global SCHEMA_NAME
    SCHEMA_NAME = schema_name

    if is_sqlite():
        prefix = f"{schema_name}."
        schema_arg = None
        return SQLModelSchemaModulePrefix, prefix, schema_arg
    else:
        prefix = ""
        schema_arg = schema_name
        return SQLModelSchemaModule, prefix, schema_arg


def get_sqlite_prefix_schema_delim_from_alembic() -> Tuple[bool, str, Optional[str], str]:
    from alembic import op

    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema, delim = f"{SCHEMA_NAME}.", None, "."
    else:
        prefix, schema, delim = "", SCHEMA_NAME, "_"

    return sqlite, prefix, schema, delim


def add_relationship_keys(table: sqm.SQLModel):  # type: ignore
    """add all relationship keys to __sqlmodel_relationships__."""
    for i in getattr(table, "__mapper__").relationships:
        if i not in getattr(table, "__sqlmodel_relationships__"):
            getattr(table, "__sqlmodel_relationships__")[i.key] = None


def add_relationship(
    origin_table: sqm.SQLModel,
    origin_attr_name: str,
    target_attr_name: Optional[str] = None,
):
    """Add relationship defined in the origin table to its respective target table.

    Only valid for many-to-many relationships.

    Args:
        origin_table (sqlmodel.SQLModel): origin (input) table
        origin_attr_name (str): name of the attribute in the origin table that defines the relationship
        target_attr_name (str, optional): name of the corresponding attribute to be added in the relationship target table.
            Defaults to either the back_populates parameter in the origin table relationship attribute (if set) or the snake-case, pluralized name of the origin table.

    Returns:
        None

    Raises:
        ValueError: if origin_attr_name does not correspond to a valid relationship attribute in the origin table
        ValueError: if target_attr_name conflicts with the back_populates parameter set in the attribute of the origin table that defines the relationship
    """
    # raise error if there is no attribute with name origin_attr_name
    origin_attr = getattr(origin_table, origin_attr_name)
    # raise error if attr is not a relationship
    relationship = getattr(origin_attr, "prop")
    if not isinstance(relationship, RelationshipProperty):
        raise ValueError("Input attribute must be a relationship.")
    # make sure relationship key is in __sqlmodel_relationships__
    if relationship not in getattr(origin_table, "__sqlmodel_relationships__"):
        getattr(origin_table, "__sqlmodel_relationships__")[relationship.key] = None
    # create relationship attribute in target table
    target_table = relationship.entity.entity
    target_attr_name = _derive_target_attr_name(origin_table, relationship, target_attr_name)
    setattr(
        target_table,
        target_attr_name,  # type: ignore
        sa_relationship(
            origin_table,
            back_populates=relationship.key,
            secondary=relationship.secondary,
        ),
    )
    # add relationship key to target table
    getattr(target_table, "__sqlmodel_relationships__")[target_attr_name] = None


def add_relationships(origin_table: sqm.SQLModel):
    """Add all relationships defined in the origin table to their corresponding target tables.

    Only valid for many-to-many relationships.

    Args:
        origin_table (sqlmodel.SQLModel): origin (input) table

    Returns:
        None
    """
    for attr in origin_table.__dict__.values():
        # skip attributes that are not relationships
        attr_property = getattr(attr, "prop", None)
        if not isinstance(attr_property, RelationshipProperty):
            continue
        relationship = attr_property
        # only configure relationships that are many-to-many
        if getattr(relationship, "secondary") is not None:
            origin_attr_name = relationship.key
            add_relationship(origin_table, origin_attr_name)


def validate_with_pydantic(model, user_kwargs):
    annotations = model.__annotations__
    kwargs = {**model.__dict__, **user_kwargs}
    pydantic_annotations = {}
    for field, ann in annotations.items():
        # ignore special fields
        if field.startswith("_"):
            kwargs.pop(field, None)
            user_kwargs.pop(field, None)
            continue

        # resolve forward references
        resolved_ann = ann
        if "ForwardRef" in str(ann):  # typing.ForwardRef not available in python <= 3.6.12
            resolved_ann = _resolve_forward_ref(ann, model.__module__)

        # handle relationship and optional fields
        if field in model.__sqlmodel_relationships__ or _is_ann_optional(ann) or _is_field_optional(model.__fields__.get(field)):
            pydantic_annotations[field] = (resolved_ann, None)
        else:
            # do not validate auto-populated (server-side) created_at fields
            if field in model.__fields__:
                sa_column_kwargs = getattr(model.__fields__[field].field_info, "sa_column_kwargs", None)
                if isinstance(sa_column_kwargs, dict):
                    if "server_default" in sa_column_kwargs and field not in user_kwargs:
                        continue
            pydantic_annotations[field] = (resolved_ann, ...)

    # validate with pydantic
    validation_model = create_model(model.__class__.__name__, **pydantic_annotations)
    validation_model.validate(kwargs)
    return


def validate_relationship_types(model, user_kwargs):
    # only validate standard sqlmodel relationships (typed) passed by the user
    standard_sqm_relationships = {k: v for k, v in model.__sqlmodel_relationships__.items() if v is not None}
    standard_user_rel_keys = standard_sqm_relationships.keys() & user_kwargs.keys()
    for key in standard_user_rel_keys:
        if "ForwardRef" in str(model.__annotations__[key]):
            rel_type = _resolve_forward_ref(model.__annotations__[key], model.__module__)
        else:
            rel_type = model.__annotations__[key]
        try:
            check_type(key, user_kwargs[key], rel_type)
        except TypeError:
            error_message = (
                f"Validation error for {model.__class__.__name__}\n"
                f"{key}\n"
                f"  should be of type {rel_type.__module__}.{rel_type}, got "
                f"{user_kwargs[key].__class__.__module__}.{user_kwargs[key].__class__.__name__} instead."
            )
            raise TypeError(error_message) from None

    # auto-populate foreign key fields associated with any relationship passed by the user (avoid missing key error during pydantic validation)
    all_user_rel_keys = model.__sqlmodel_relationships__.keys() & user_kwargs.keys()
    for relationship_name in all_user_rel_keys:
        rel_local_fields = getattr(model.__class__, relationship_name).property.local_columns  # model fields associated with the relationship
        rel_fk_fields = [field for field in rel_local_fields if not field.primary_key]  # primary keys are returned as local fields in link table relationships
        rel_fk_field_names = [field.name for field in rel_fk_fields]
        target_col_names = [list(field.foreign_keys)[0].column.name for field in rel_fk_fields]
        for field_name, col_name in zip(rel_fk_field_names, target_col_names):
            if hasattr(model, field_name):
                if getattr(model, field_name) is None:
                    try:
                        setattr(model, field_name, getattr(user_kwargs[relationship_name], col_name))
                    except AttributeError:
                        pass


def _derive_target_attr_name(
    origin_table: sqm.SQLModel,
    origin_relationship: RelationshipProperty,
    custom_target_name,
):
    """Derive the name of the relationship attribute in the target table."""
    # If set, back_populates parameter in the origin relationship takes precedence over any custom naming
    if origin_relationship.back_populates is not None:
        if custom_target_name is not None and custom_target_name != origin_relationship.back_populates:
            raise ValueError(
                "Custom name for the relationship attribute in the target table conflicts with back_populates parameter set in the relationship attribute of the origin table."
            )
        return origin_relationship.back_populates
    else:
        if custom_target_name is not None:
            return custom_target_name
        else:
            # Convert origin table name from camel to snake case and make it plural
            return f"{re.sub(r'(?<!^)(?=[A-Z])', '_', origin_table.__name__).lower()}s"


def _is_ann_optional(type_ann):
    if type_ann.__module__ == "typing":
        try:
            if type_ann.__origin__ == Union:
                if type(None) in type_ann.__args__:
                    return True
        except Exception:
            return False
    return False


def _is_field_optional(model_field):
    if model_field is None:
        return False
    return not model_field.required


def _resolve_forward_ref(ann, module):
    # fetch target type as string
    forward_args = []
    parent_ref = None
    if _is_ann_optional(ann):
        forward_args = [ann.__args__[0].__forward_arg__]
    else:
        try:
            for arg in ann.__args__:
                if "ForwardRef" in str(arg):
                    forward_args += [arg.__forward_arg__]
            parent_ref = ann._name
        except Exception:
            forward_args = [ann.__forward_arg__]

    # fetch target type from string
    resolved_args = []
    for arg in forward_args:
        split_arg = arg.split(".")
        try:
            # if reference is in another module
            resolved_args += [getattr(importlib.import_module(split_arg[0]), split_arg[-1])]
        except Exception:
            resolved_args += [getattr(importlib.import_module(module), split_arg[-1])]

    # return resolved annotation
    if len(resolved_args) == 1:
        if parent_ref is None:
            return resolved_args[0]
        else:
            return getattr(typing, parent_ref)[resolved_args[0]]
    else:
        # This code should never run. SQLmodel is incompatible with non-optional
        # Union-like types (e.g. typing.Union[str, ForwardRef("lnschema_core._core.Transform")])
        # and raises an error at class definition when a field is typed as such.
        return getattr(typing, parent_ref)[resolved_args[0], resolved_args[1]]


# # as a decorator
# def db_lookup(lookup_field: str = 'name'):
#     def wrapper(sqlmodel_class):
#         @classproperty
#         def df(cls):
#             import lamindb as ln

#             return ln.select(sqlmodel_class).df()

#         @classproperty
#         def lookup(cls):
#             return lookup_field

#         sqlmodel_class.df = df
#         sqlmodel_class.lookup = lookup
#         return sqlmodel_class

#     return wrapper


def _to_lookup_keys(x: Iterable[str], padding: str = "LOOKUP") -> list:
    """Convert a list of strings to tab-completion allowed formats."""
    lookup = [re.sub("[^0-9a-zA-Z]+", "_", str(i)) for i in x]
    for i, value in enumerate(lookup):
        if value == "" or (not value[0].isalpha()):
            lookup[i] = f"{padding}_{value}"
    return lookup


def _namedtuple_from_dict(d: dict, name: str = "entity") -> tuple:
    """Create a namedtuple from a dict to allow autocompletion."""
    nt = namedtuple(name, d.keys())  # type:ignore
    return nt(**d)



"""v0.24.0.

Revision ID: f6b6b85cdffc
Revises: db1df7b2aaad
Create Date: 2023-01-09 12:05:35.536833

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f6b6b85cdffc"
down_revision = "db1df7b2aaad"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    if sqlite:
        with op.batch_alter_table(f"{prefix}notebook", schema=schema) as batch_op:
            batch_op.add_column(sa.Column("title", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
            batch_op.create_index(batch_op.f("ix_core.notebook_title"), ["title"], unique=False)
    else:
        op.add_column(f"{prefix}notebook", sa.Column("title", sqlmodel.sql.sqltypes.AutoString(), nullable=True), schema=schema)
        op.create_index(op.f("ix_core_notebook_title"), "notebook", ["title"], unique=False, schema=schema)


def downgrade() -> None:
    pass



"""v0.12.0.

Revision ID: 2ddcb037e3ea
Revises: 3d244a8d3148
Create Date: 2022-10-11 17:33:37.731501

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2ddcb037e3ea"
down_revision = "3d244a8d3148"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("dobject", schema=None) as batch_op:
        batch_op.add_column(sa.Column("checksum", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
        batch_op.create_index(batch_op.f("ix_dobject_checksum"), ["checksum"], unique=False)


def downgrade() -> None:
    pass



from datetime import datetime as datetime
from pathlib import Path, PurePosixPath
from typing import Any, List, Optional, TypeVar, Union, overload  # noqa

import anndata as ad
import pandas as pd
import sqlalchemy as sa
from lamin_logger import logger
from lndb.dev.upath import UPath
from nbproject._is_run_from_ipython import is_run_from_ipython
from pydantic.fields import PrivateAttr
from sqlmodel import Field, ForeignKeyConstraint, Relationship

from . import _name as schema_name
from ._link import FileFeatures, FolderFile, ProjectFolder, RunIn  # noqa
from ._timestamps import CreatedAt, UpdatedAt
from ._users import CreatedBy
from .dev import id as idg
from .dev.sqlmodel import schema_sqlmodel
from .types import DataLike, PathLike, TransformType

SQLModel, prefix, schema_arg = schema_sqlmodel(schema_name)


class User(SQLModel, table=True):  # type: ignore
    """User accounts.

    All data in this table is synched from the cloud user account to ensure a
    globally unique user identity.
    """

    id: str = Field(primary_key=True)
    email: str = Field(index=True, unique=True)
    """Primary user email used for logging in."""
    handle: str = Field(nullable=False, index=True, unique=True)
    """User handle."""
    name: Optional[str] = Field(index=True)
    """Long display name."""
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class Storage(SQLModel, table=True):  # type: ignore
    """Storage locations.

    A file or run-associated file can be stored in any desired S3,
    GCP, Azure or local storage location. This table tracks these locations
    along with metadata.
    """

    id: str = Field(default_factory=idg.storage, primary_key=True)
    root: str = Field(index=True)
    """Semantic identifier to the root of the storage location, like an s3 path, a local path, etc."""  # noqa
    type: Optional[str] = None
    """Local vs. s3 vs. gcp etc."""
    region: Optional[str] = None
    """Cloud storage region if applicable."""
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt
    created_by: User = Relationship()
    created_by_id: Optional[str] = CreatedBy  # make non-optional over time


class Project(SQLModel, table=True):  # type: ignore
    """Projects."""

    id: str = Field(default_factory=idg.project, primary_key=True)
    name: str = Field(index=True)
    created_by: User = Relationship()
    created_by_id: str = CreatedBy
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class Transform(SQLModel, table=True):  # type: ignore
    """Data transformations.

    Jupyter notebooks, pipelines, and apps.

    A pipeline is typically versioned software that can perform a data
    transformation/processing workflow. This can be anything from typical
    workflow tools (Nextflow, Snakemake, Prefect, Apache Airflow, etc.) to
    simple (versioned) scripts.

    Data can also be ingested & transformed through an app.
    """

    id: Optional[str] = Field(sa_column=sa.Column(sa.String, primary_key=True, default=idg.transform))
    version: Optional[str] = Field(sa_column=sa.Column(sa.String, primary_key=True, default="0"))
    """Version identifier, defaults to `"1"`.

    Use this to label different versions of the same transform.

    Consider using `semantic versioning <https://semver.org>`__
    with `Python versioning <https://peps.python.org/pep-0440/>`__.
    """
    name: str = Field(index=True)
    """A name for the transform, a pipeline name, or a file name of a notebook or script.
    """
    type: TransformType = Field(
        index=True,
        default=(TransformType.notebook if is_run_from_ipython else TransformType.pipeline),
    )
    """Transform type: defaults to `notebook` if run from IPython and otherwise to `pipeline`.
    """
    title: Optional[str] = Field(index=True)
    """An additional title, like a notebook title.
    """
    reference: Optional[str] = Field(index=True)
    """Reference for the transform, e.g., a URL.
    """
    created_by: User = Relationship()
    created_by_id: str = CreatedBy
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class Run(SQLModel, table=True):  # type: ignore
    """Runs of data transforms.

    A `run` is any transform of a `file`.

    Args:
        id: Optional[str] = None
        name: Optional[str] = None
        load_latest: bool = False - Load latest run for given notebook or pipeline.
        transform: Optional[Transform] = None
        inputs: List[File] = None
        outputs: List[File] = None

    It typically has inputs and outputs:

    - References to outputs are stored in the `file` table in the
      `run_id` column as a foreign key the `run`
      table. This is possible as every given `file` has a unique data run:
      the `run` that produced the `file`. However, note that a given
      `run` may output several `files`.
    - References to inputs are stored in the `run_in` table, a
      many-to-many link table between the `file` and `run` tables. Any
      `file` might serve as an input for many `runs`. Similarly, any
      `run` might have many `files` as inputs.
    """

    __table_args__ = (
        ForeignKeyConstraint(
            ["transform_id", "transform_version"],
            ["core.transform.id", "core.transform.version"],
        ),
        {"schema": schema_arg},
    )
    id: Optional[str] = Field(default_factory=idg.run, primary_key=True)
    name: Optional[str] = Field(default=None, index=True)
    external_id: Optional[str] = Field(default=None, index=True)
    transform_id: Optional[str] = Field(default=None, index=True)
    transform_version: Optional[str] = Field(default=None, index=True)
    transform: Transform = Relationship()
    outputs: List["File"] = Relationship(back_populates="run")
    inputs: List["File"] = Relationship(
        back_populates="input_of",
        sa_relationship_kwargs=dict(secondary=RunIn.__table__),
    )
    created_by: User = Relationship()
    created_by_id: str = CreatedBy
    created_at: datetime = CreatedAt

    _ln_identity_key: Optional[str] = PrivateAttr(default=None)
    # simulate query result

    def __init__(  # type: ignore
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        load_latest: bool = False,
        external_id: Optional[str] = None,
        transform: Optional[Transform] = None,
        inputs: List["File"] = None,
        outputs: List["File"] = None,
    ):
        kwargs = {k: v for k, v in locals().items() if v and k != "self"}

        import lamindb as ln

        global_context = False
        if transform is None:
            if ln.context.transform is not None:
                global_context = True
                transform = ln.context.transform
            else:
                raise ValueError("Either call `ln.Run(transform=transform)` or `ln.track(transform=...)`.")

        if not isinstance(transform, Transform):
            raise TypeError("transform needs to be of type Transform")

        run = None
        if load_latest:
            run = (
                ln.select(
                    ln.Run,
                    transform_id=transform.id,
                    transform_version=transform.version,
                )
                .order_by(ln.Run.created_at.desc())
                .first()
            )
            if run is not None:
                logger.info(f"Loaded: {run}")
        elif id is not None:
            run = ln.select(ln.Run, id=id).one_or_none()
            if run is None:
                raise NotImplementedError("You can currently only pass existing ids")

        if run is None:
            kwargs.update(dict(transform_id=transform.id, transform_version=transform.version))
            super().__init__(**kwargs)
            self._ln_identity_key = None
        else:
            super().__init__(**run.dict())
            self._ln_identity_key = run.id  # simulate query result

        if global_context:
            if run is None:
                added_self = ln.add(self)
                self._ln_identity_key = added_self.id
                logger.success(f"Added: {self}")
            ln.context.run = self


class Features(SQLModel, table=True):  # type: ignore
    """Feature sets.

    A feature set is represented by the hash of the set of primary keys and the feature type.

    The current supported feature types are lnschema_bionty.Gene,
    lnschema_bionty.Protein & lnschema_bionty.CellMarker.

    Guides:

    - :doc:`/guide/scrna`
    - :doc:`guide/flow`

    Examples:

    >>> import lnschema_bionty as bt
    >>> reference = bt.Gene(species="mouse")
    >>> features = ln.Features(adata, reference=reference)
    >>> file = ln.File(adata, name="Mouse Lymph Node scRNA-seq", features=features)

    Args:
        data: [Path, str, pd.DataFrame, ad.AnnData] - DataFrame or AnnData to parse.
        reference: Any = None - Reference for mapping features.
        id: str = None - Primary key.
        type: Any = None - Type of reference.
        files: List[File] - Files to link against.
    """

    id: str = Field(primary_key=True)  # use a hash
    type: str  # was called entity_type
    created_by: User = Relationship()
    created_by_id: str = CreatedBy
    created_at: datetime = CreatedAt
    files: List["File"] = Relationship(
        back_populates="features",
        sa_relationship_kwargs=dict(secondary=FileFeatures.__table__),
    )

    @overload
    def __init__(
        self,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        reference: Any = None,
    ):
        """Initialize from data."""
        ...

    @overload
    def __init__(
        self,
        id: Optional[str] = None,
        type: Optional[str] = None,
        files: List["File"] = [],
    ):
        """Initialize from fields."""
        ...

    def __init__(  # type: ignore
        self,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        reference: Any = None,
        *,
        id: str = None,
        type: Any = None,
        # continue with fields
        files: List["File"] = [],
    ):
        kwargs = {k: v for k, v in locals().items() if v and k != "self"}
        super().__init__(**kwargs)

    def __new__(
        cls,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        reference: Any = None,
        *,
        id: str = None,
        type: Any = None,
        # continue with fields
        files: List["File"] = [],
    ):
        if data is not None:
            from lamindb._file import get_features_from_data

            features = get_features_from_data(
                data=data,
                reference=reference,
            )
        else:
            features = super().__new__(cls)
        return features


class Folder(SQLModel, table=True):  # type: ignore
    """See lamindb for docstring."""

    __table_args__ = (
        sa.UniqueConstraint("storage_id", "key", name="uq_folder_storage_key"),
        {"schema": schema_arg},
    )

    id: str = Field(default_factory=idg.folder, primary_key=True)
    name: str = Field(index=True)
    key: Optional[str] = Field(default=None, index=True)
    storage_id: Optional[str] = Field(default=None, foreign_key="core.storage.id", index=True)
    """Storage root id."""
    files: List["File"] = Relationship(  # type: ignore  # noqa
        back_populates="folders",
        sa_relationship_kwargs=dict(secondary=FolderFile.__table__),
    )
    """:class:`~lamindb.File`."""
    created_by: User = Relationship()
    created_by_id: str = CreatedBy
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt

    # private attributes are needed here to prevent sqlalchemy error
    _local_filepath: Optional[Path] = PrivateAttr()
    _cloud_filepath: Optional[UPath] = PrivateAttr()

    @property
    def __name__(cls) -> str:
        return "Folder"

    def path(self) -> Union[Path, UPath]:
        """Path on storage."""
        return filepath_from_file_or_folder(self)

    def tree(
        self,
        level: int = -1,
        limit_to_directories: bool = False,
        length_limit: int = 1000,
    ) -> None:
        """Print a visual tree structure."""
        from lamindb._folder import tree

        return tree(
            dir_path=self.path(),
            level=level,
            limit_to_directories=limit_to_directories,
            length_limit=length_limit,
        )

    @overload
    def __init__(
        self,
        path: Union[Path, UPath, str] = None,
        *,
        name: Optional[str] = None,
    ):
        """Initialize from folder."""
        ...

    @overload
    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        files: List["File"] = [],
    ):
        """Initialize from fields."""
        ...

    def __init__(  # type: ignore
        self,
        path: Optional[Union[Path, UPath, str]] = None,
        *,
        # continue with fields
        id: Optional[str] = None,
        name: Optional[str] = None,
        key: Optional[str] = None,
        storage_id: Optional[str] = None,
        files: List["File"] = [],
    ):
        if path is not None:
            from lamindb._folder import get_folder_kwargs_from_data

            kwargs, privates = get_folder_kwargs_from_data(
                path=path,
                name=name,
                key=key,
            )
            if id is not None:
                kwargs["id"] = id
        else:
            kwargs = {k: v for k, v in locals().items() if v and k != "self"}

        super().__init__(**kwargs)
        if path is not None:
            self._local_filepath = privates["local_filepath"]
            self._cloud_filepath = privates["cloud_filepath"]


class File(SQLModel, table=True):  # type: ignore
    """See lamindb for docstring."""

    __table_args__ = (
        sa.UniqueConstraint("storage_id", "key", name="uq_file_storage_key"),
        ForeignKeyConstraint(
            ["transform_id", "transform_version"],
            ["core.transform.id", "core.transform.version"],
            name="fk_file_transform_id_version_transform",
        ),
        {"schema": schema_arg},
    )

    id: str = Field(default_factory=idg.file, primary_key=True)
    name: Optional[str] = Field(index=True)
    suffix: Optional[str] = Field(default=None, index=True)
    """Suffix to construct the storage key. Defaults to `None`.

    This is a file extension if the `file` is stored in a file format.
    It's `None` if the storage format doesn't have a canonical extension.
    """

    size: Optional[int] = Field(default=None, sa_column=sa.Column(sa.BigInteger(), index=True))
    """Size in bytes.

    Examples: 1KB is 1e3 bytes, 1MB is 1e6, 1GB is 1e9, 1TB is 1e12 etc.
    """
    hash: Optional[str] = Field(default=None, index=True)
    """Hash (md5)."""
    key: Optional[str] = Field(default=None, index=True)
    """Storage key: relative path within storage location."""
    run: Optional[Run] = Relationship(back_populates="outputs")  # type: ignore
    """:class:`~lamindb.Run` that generated the `file`."""
    run_id: Optional[str] = Field(foreign_key="core.run.id", index=True)
    """Source run id."""
    transform: Transform = Relationship()  # type: ignore
    """:class:`~lamindb.Transform` whose run generated the `file`."""
    transform_id: Optional[str] = Field(index=True)
    """Source transform id."""
    transform_version: Optional[str] = Field(index=True)
    """Source transform version."""
    storage: Storage = Relationship()  # type: ignore
    """:class:`~lamindb.Storage` location of `file`, see `.path()` for full path."""
    storage_id: str = Field(foreign_key="core.storage.id", index=True)
    """Storage root id."""
    features: List[Features] = Relationship(
        back_populates="files",
        sa_relationship_kwargs=dict(secondary=FileFeatures.__table__),
    )
    """Feature sets indexing this file."""
    folders: List[Folder] = Relationship(
        back_populates="files",
        sa_relationship_kwargs=dict(secondary=FolderFile.__table__),
    )
    """Folders that contain this file."""
    input_of: List[Run] = Relationship(  # type: ignore
        back_populates="inputs",
        sa_relationship_kwargs=dict(secondary=RunIn.__table__),
    )
    """Runs that use this file as input."""
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt
    created_by: User = Relationship()
    created_by_id: Optional[str] = CreatedBy  # make non-optional over time

    # private attributes are needed here to prevent sqlalchemy error
    _local_filepath: Optional[Path] = PrivateAttr()
    _cloud_filepath: Optional[UPath] = PrivateAttr()
    _clear_storagekey: Optional[str] = PrivateAttr()
    _memory_rep: Any = PrivateAttr()
    _to_store: bool = PrivateAttr()  # indicate whether upload needed

    def path(self) -> Union[Path, UPath]:
        """Path on storage."""
        return filepath_from_file_or_folder(self)

    # likely needs an arg `key`
    def replace(
        self,
        data: Union[PathLike, DataLike],
        run: Optional[Run] = None,
        format: Optional[str] = None,
    ) -> None:
        """Replace data object."""
        from lamindb._file import get_file_kwargs_from_data

        if isinstance(data, (Path, str)):
            name_to_pass = None
        else:
            name_to_pass = self.name

        kwargs, privates = get_file_kwargs_from_data(
            data=data,
            name=name_to_pass,
            run=run,
            format=format,
        )

        if kwargs["name"] != self.name:
            logger.warning(f"Your new filename '{kwargs['name']}' does not match the previous filename '{self.name}': to update the name, set file.name = '{kwargs['name']}'")

        if self.key is not None:
            key_path = PurePosixPath(self.key)
            if isinstance(data, (Path, str)):
                new_name = kwargs["name"]  # use the name from the data filepath
            else:
                # do not change the key stem to file.name
                new_name = key_path.stem  # use the stem of the key for in-memory data
            if PurePosixPath(new_name).suffixes == []:
                new_name = f"{new_name}{kwargs['suffix']}"
            if key_path.name != new_name:
                self._clear_storagekey = self.key
                self.key = str(key_path.with_name(new_name))
                logger.warning(f"Replacing the file will also replace the key from '{key_path}' to '{self.key}', and delete '{key_path}' upon `ln.add`")
        else:
            self.key = kwargs["key"]
            old_storage = f"{self.id}{self.suffix}"
            new_storage = self.key if self.key is not None else f"{self.id}{kwargs['suffix']}"
            if old_storage != new_storage:
                self._clear_storagekey = old_storage

        self.suffix = kwargs["suffix"]
        self.size = kwargs["size"]
        self.hash = kwargs["hash"]
        self.run = kwargs["run"]
        self._local_filepath = privates["local_filepath"]
        self._cloud_filepath = privates["cloud_filepath"]
        self._memory_rep = privates["memory_rep"]
        self._to_store = not privates["check_path_in_storage"]  # no need to upload if new file is already in storage

    @property
    def __name__(cls) -> str:
        return "File"

    def __init__(  # type: ignore
        self,
        data: Union[PathLike, DataLike] = None,
        *,
        key: Optional[str] = None,
        name: Optional[str] = None,
        run: Optional[Run] = None,
        format: Optional[str] = None,
        features: List[Features] = None,
        id: Optional[str] = None,
        input_of: List[Run] = None,
    ):
        if features is None:
            features = []
        if input_of is None:
            input_of = []
        if not isinstance(features, List):
            features = [features]

        def log_hint(*, check_path_in_storage: bool, key: str, id: str, suffix: str) -> None:
            hint = ""
            if check_path_in_storage:
                hint += "file in storage ✓"
            else:
                hint += "file will be copied to storage upon `ln.add()`"
            if key is None:
                hint += f" using storage key = {id}{suffix}"
            else:
                hint += f" using storage key = {key}"
            logger.hint(hint)

        from lamindb._file import get_file_kwargs_from_data

        kwargs, privates = get_file_kwargs_from_data(
            data=data,
            name=name,
            key=key,
            run=run,
            format=format,
        )
        kwargs["id"] = idg.file() if id is None else id
        if features is not None:
            kwargs["features"] = features
        log_hint(
            check_path_in_storage=privates["check_path_in_storage"],
            key=kwargs["key"],
            id=kwargs["id"],
            suffix=kwargs["suffix"],
        )

        # transform cannot be directly passed, just via run
        # it's directly stored in the file table to avoid another join
        # mediate by the run table
        if kwargs["run"] is not None:
            if kwargs["run"].transform_id is not None:
                kwargs["transform_id"] = kwargs["run"].transform_id
                assert kwargs["run"].transform_version is not None
                kwargs["transform_version"] = kwargs["run"].transform_version
            else:
                # accessing the relationship should always be possible if
                # the above if clause was false as then, we should have a fresh
                # Transform object that is not queried from the DB
                assert kwargs["run"].transform is not None
                kwargs["transform"] = kwargs["run"].transform

        super().__init__(**kwargs)
        if data is not None:
            self._local_filepath = privates["local_filepath"]
            self._cloud_filepath = privates["cloud_filepath"]
            self._memory_rep = privates["memory_rep"]
            self._to_store = not privates["check_path_in_storage"]


# add type annotations back asap when re-organizing the module
def storage_key_from_file(file: File):
    if file.key is None:
        return f"{file.id}{file.suffix}"
    else:
        return file.key


# add type annotations back asap when re-organizing the module
def filepath_from_file_or_folder(file_or_folder: Union[File, Folder]):
    from lndb import settings
    from lndb.dev import StorageSettings

    # using __name__ for type check to avoid need of
    # dynamically importing the type
    if file_or_folder.__name__ == "File":
        storage_key = storage_key_from_file(file_or_folder)
    else:
        storage_key = file_or_folder.key
        if storage_key is None:
            raise ValueError("Only real folders have a path!")
    if file_or_folder.storage_id == settings.storage.id:
        path = settings.storage.key_to_filepath(storage_key)
    else:
        logger.warning(
            "file.path() is slow for files outside the currently configured storage"
            " location\nconsider joining for the set of files you're interested in:"
            " ln.select(ln.File, ln.Storage)the path is storage.root / file.key if"
            " file.key is not None\notherwise storage.root / (file.id + file.suffix)"
        )
        import lamindb as ln

        storage = ln.select(ln.Storage, id=file_or_folder.storage_id).one()
        # find a better way than passing None to instance_settings in the future!
        storage_settings = StorageSettings(storage.root, instance_settings=None)
        path = storage_settings.key_to_filepath(storage_key)
    return path



"""v0.22.0.

Revision ID: db1df7b2aaad
Revises: 9d94f3b9566d
Create Date: 2022-12-07 01:40:35.071647

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "db1df7b2aaad"
down_revision = "9d94f3b9566d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    op.rename_table(f"{prefix}jupynb", f"{prefix}notebook", schema=schema)

    with op.batch_alter_table(f"{prefix}notebook", schema=schema) as batch_op:
        batch_op.create_index(batch_op.f("ix_core.notebook_created_at"), ["created_at"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.notebook_created_by"), ["created_by"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.notebook_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.notebook_updated_at"), ["updated_at"], unique=False)
        batch_op.drop_index("ix_core.jupynb_created_at")
        batch_op.drop_index("ix_core.jupynb_created_by")
        batch_op.drop_index("ix_core.jupynb_name")
        batch_op.drop_index("ix_core.jupynb_updated_at")

    op.alter_column(
        f"{prefix}run",
        column_name="jupynb_id",
        new_column_name="notebook_id",
        schema=schema,
    )
    op.alter_column(
        f"{prefix}run",
        column_name="jupynb_v",
        new_column_name="notebook_v",
        schema=schema,
    )
    with op.batch_alter_table(f"{prefix}run", schema=schema) as batch_op:
        batch_op.drop_index("ix_core.run_jupynb_id")
        batch_op.drop_index("ix_core.run_jupynb_v")
        batch_op.create_index(batch_op.f("ix_core.run_notebook_id"), ["notebook_id"], unique=False)
        batch_op.create_index(batch_op.f("ix_core.run_notebook_v"), ["notebook_v"], unique=False)
        batch_op.create_foreign_key(
            batch_op.f("fk_core.run_notebook_id_notebook"),
            "core.notebook",
            ["notebook_id", "notebook_v"],
            ["id", "v"],
        )


def downgrade() -> None:
    pass



import numpy as np
import pandas as pd


def biosample():
    """Simple, dummy biosample example."""
    sample = pd.DataFrame(
        {
            "Name": [
                "hc_hea_021",
                "cv_acu_027",
                "cv_acu_010",
                "cv_con_049",
                "cv_rec_002",
            ],
            "Species": ["human", "human", "human", "human", "human"],
            "Cell Type": ["CD8+T", "CD8+T", "CD8+T", "CD8+T", "CD8+T"],
            "Experiment": ["001", "002", "003", "004", "004"],
            "Donor": ["021", "027", "010", "049", "002"],
            "Disease": [None, "U07.1", "U07.1", "U07.1", "U07.1"],
            "Custom 1": ["healthy", "acute", "acute", "convalescent", "recovered"],
            "Custom 2": ["control", "covid-19", "covid-19", "covid-19", "covid-19"],
            "Custom 3": [12.11, np.nan, 0.87, np.nan, 11.91],
        }
    )
    return sample


def techsample():
    """Simple, dummy techsample example."""
    sample = pd.DataFrame(
        {
            "Name": ["TS001", "TS002", "TS003", "TS004", "TS005"],
            "Batch": [1, 1, 1, 2, 2],
            "File Type": ["fastq", "fastq", "fastq", "fastq", "fastq"],
            "Filepath R1": [
                "SRX1603629_T1_1.fastq.gz",
                "SRX1603629_T1_2.fastq.gz",
                "SRX1603629_T1_3.fastq.gz",
                "SRX1603629_T1_4.fastq.gz",
                "SRX1603629_T1_5.fastq.gz",
            ],
            "Filepath R2": [
                "SRX1603629_T2_1.fastq.gz",
                "SRX1603629_T2_2.fastq.gz",
                "SRX1603629_T2_3.fastq.gz",
                "SRX1603629_T2_4.fastq.gz",
                "SRX1603629_T2_5.fastq.gz",
            ],
            "Custom 1": ["13.42%", "2.43%", "4.57%", np.nan, "9.36%"],
            "Custom 2": [
                "Gene Expression",
                "Gene Expression",
                "Gene Expression",
                "Gene Expression",
                "Gene Expression",
            ],
        }
    )
    return sample



# Some IDs got updated on 2023-02-14 to be multiples of 4
# for matching byte representation, should we ever go there.
import secrets
import string


def base64(n_char: int) -> str:
    """Like nanoid.

    Alphabet: `string.digits + string.ascii_letters.swapcase() + "_-"`
    """
    alphabet = string.digits + string.ascii_letters.swapcase() + "_" + "-"
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def base62(n_char: int) -> str:
    """Like nanoid without hyphen and underscore."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def base26(n_char: int):
    """ASCII lowercase."""
    alphabet = string.ascii_lowercase
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def schema():
    """Schema module: 4 base26."""
    return base26(4)


def file() -> str:
    """Data object: 20 base62.

    20 characters (62**20=7e+35 possibilities) roughly matches UUID (2*122=5e+36).

    Can be encoded as 15 bytes.
    """
    return base62(n_char=20)


def folder() -> str:
    """Data folder: 20 base62."""
    return base62(n_char=20)


def run() -> str:
    """Data transformation: 20 base62."""
    return base62(n_char=20)


def user() -> str:
    """User: 8 base62.

    Consistent with 1M users producing 1k notebooks.
    Safe for 100k users producing 10k notebooks.

    Allows >2e14 users.

    This is one of 2 IDs that are centralized.

    Collision probability in decentralized system is:

    ======= ===========
    n_users p_collision
    ======= ===========
    100k    2e-05
    1M      2e-03
    ======= ===========
    """
    return base62(n_char=8)


def project() -> str:
    """Project: 8 base62."""
    return base62(n_char=8)


def usage() -> str:
    """Usage event: 24 base62."""
    return base62(n_char=24)


def secret() -> str:
    """Password or secret: 40 base62."""
    return base62(n_char=40)


def instance() -> str:
    """LaminDB instance: 12 base62.

    Collision probability is 2e-04 for 1B instances: 1M users with 1k instances/user.
    """
    return base62(n_char=12)


def storage() -> str:
    """Storage root: 8 base62."""
    return base62(n_char=8)


def transform():
    """Transforms: 12 base62.

    Collision probability is 2e-04 for 1B transforms: 1M users with 1k transforms/user.

    Is the same as nbproject ID!
    """
    return base62(n_char=12)



"""v0.9.0.

Revision ID: 439c4ee0a22a
Revises: 1190648443cb
Create Date: 2022-09-30 08:13:39.190602

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "439c4ee0a22a"
down_revision = "1190648443cb"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("dobject", schema=None) as batch_op:
        batch_op.drop_column("v")

    with op.batch_alter_table("dtransform_in", schema=None) as batch_op:
        batch_op.drop_constraint("dtransform_in_dobject", type_="foreignkey")
        batch_op.create_foreign_key("fk_dobject_id", "dobject", ["dobject_id"], ["id"])
        batch_op.drop_column("dobject_v")

    with op.batch_alter_table("usage", schema=None) as batch_op:
        batch_op.drop_index("ix_usage_dobject_v")
        batch_op.drop_constraint("usage_dobject", type_="foreignkey")
        batch_op.create_foreign_key("fk_dobject_id", "dobject", ["dobject_id"], ["id"])
        batch_op.drop_column("dobject_v")


def downgrade() -> None:
    pass



"""v0.30rc3."""
import sqlalchemy as sa  # noqa
import sqlmodel as sqm  # noqa
from alembic import op
from sqlalchemy.engine.reflection import Inspector

from lnschema_core.dev.sqlmodel import get_sqlite_prefix_schema_delim_from_alembic

revision = "5846a15d9241"
down_revision = "ebafd37fd6e1"


def upgrade() -> None:
    engine = op.get_bind().engine
    inspector = Inspector.from_engine(engine)

    sqlite, prefix, schema, delim = get_sqlite_prefix_schema_delim_from_alembic()

    op.rename_table(old_table_name=f"{prefix}dobject", new_table_name=f"{prefix}file", schema=schema)
    op.rename_table(old_table_name=f"{prefix}dfolder", new_table_name=f"{prefix}folder", schema=schema)
    op.rename_table(old_table_name=f"{prefix}project_dfolder", new_table_name=f"{prefix}project_folder", schema=schema)
    op.rename_table(old_table_name=f"{prefix}dfolder_dobject", new_table_name=f"{prefix}folder_file", schema=schema)
    op.rename_table(old_table_name=f"{prefix}dobject_features", new_table_name=f"{prefix}file_features", schema=schema)

    try:
        op.drop_index(f"ix_core{delim}dobject__objectkey", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_created_at", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_hash", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_name", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_size", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_source_id", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_storage_id", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_suffix", table_name=f"{prefix}file", schema=schema)
        op.drop_index(f"ix_core{delim}dobject_updated_at", table_name=f"{prefix}file", schema=schema)
    except Exception:
        pass
    op.create_index(op.f(f"ix_core{delim}file__objectkey"), f"{prefix}file", ["_objectkey"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_created_at"), f"{prefix}file", ["created_at"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_hash"), f"{prefix}file", ["hash"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_name"), f"{prefix}file", ["name"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_size"), f"{prefix}file", ["size"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_source_id"), f"{prefix}file", ["source_id"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_storage_id"), f"{prefix}file", ["storage_id"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_suffix"), f"{prefix}file", ["suffix"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}file_updated_at"), f"{prefix}file", ["updated_at"], unique=False, schema=schema)
    try:
        op.drop_index(f"ix_core{delim}dfolder__objectkey", table_name=f"{prefix}folder", schema=schema)
        op.drop_index(f"ix_core{delim}dfolder_created_at", table_name=f"{prefix}folder", schema=schema)
        op.drop_index(f"ix_core{delim}dfolder_created_by", table_name=f"{prefix}folder", schema=schema)
        op.drop_index(f"ix_core{delim}dfolder_name", table_name=f"{prefix}folder", schema=schema)
        op.drop_index(f"ix_core{delim}dfolder_updated_at", table_name=f"{prefix}folder", schema=schema)
    except Exception:
        pass
    op.create_index(op.f(f"ix_core{delim}folder__objectkey"), f"{prefix}folder", ["_objectkey"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}folder_created_at"), f"{prefix}folder", ["created_at"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}folder_created_by"), f"{prefix}folder", ["created_by"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}folder_name"), f"{prefix}folder", ["name"], unique=False, schema=schema)
    op.create_index(op.f(f"ix_core{delim}folder_updated_at"), f"{prefix}folder", ["updated_at"], unique=False, schema=schema)

    with op.batch_alter_table(f"{prefix}folder_file", schema=schema) as batch_op:
        batch_op.alter_column(column_name="dobject_id", new_column_name="file_id")
        batch_op.alter_column(column_name="dfolder_id", new_column_name="folder_id")
        for constraint in inspector.get_foreign_keys(f"{prefix}dfolder_dobject", schema=schema):
            if constraint["name"] == "fk_dfolder_dobject_dobject_id_dobject":
                batch_op.drop_constraint("fk_dfolder_dobject_dobject_id_dobject", type_="foreignkey")
            if constraint["name"] == "fk_dfolder_dobject_dfolder_id_dfolder":
                batch_op.drop_constraint("fk_dfolder_dobject_dfolder_id_dfolder", type_="foreignkey")
        batch_op.create_foreign_key(op.f("fk_folder_file_file_id_file"), f"{prefix}file", ["file_id"], ["id"], referent_schema=schema)
        batch_op.create_foreign_key(op.f("fk_folder_file_folder_id_folder"), f"{prefix}folder", ["folder_id"], ["id"], referent_schema=schema)

    with op.batch_alter_table(f"{prefix}file_features", schema=schema) as batch_op:
        batch_op.alter_column(column_name="dobject_id", new_column_name="file_id")
        for constraint in inspector.get_foreign_keys(f"{prefix}dobject_features", schema=schema):
            if constraint["name"] == "fk_dobject_features_dobject_id_dobject":
                batch_op.drop_constraint("fk_dobject_features_dobject_id_dobject", type_="foreignkey")
        batch_op.create_foreign_key(op.f("fk_file_features_file_id_file"), f"{prefix}file", ["file_id"], ["id"], referent_schema=schema)

    with op.batch_alter_table(f"{prefix}project_folder", schema=schema) as batch_op:
        batch_op.alter_column(column_name="dfolder_id", new_column_name="folder_id")
        for constraint in inspector.get_foreign_keys(f"{prefix}project_dfolder", schema=schema):
            if constraint["name"] == "fk_project_dfolder_dfolder_id_dfolder":
                batch_op.drop_constraint("fk_project_dfolder_dfolder_id_dfolder", type_="foreignkey")
        batch_op.create_foreign_key(op.f("fk_project_folder_folder_id_folder"), f"{prefix}folder", ["folder_id"], ["id"], referent_schema=schema)

    with op.batch_alter_table(f"{prefix}run_in", schema=schema) as batch_op:
        batch_op.alter_column(column_name="dobject_id", new_column_name="file_id")
        for constraint in inspector.get_foreign_keys(f"{prefix}run_in", schema=schema):
            if constraint["name"] == "fk_run_in_dobject_id_dobject":
                batch_op.drop_constraint("fk_run_in_dobject_id_dobject", type_="foreignkey")
        batch_op.create_foreign_key(op.f("fk_run_in_file_id_file"), f"{prefix}file", ["file_id"], ["id"], referent_schema=schema)


def downgrade() -> None:
    pass



from pathlib import Path
from typing import Union
from urllib.request import urlretrieve

import anndata as ad
import numpy as np
import pandas as pd
from lnschema_core.dev import id


def file_fcs() -> Path:
    """Return fcs file example."""
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/example.fcs", "example.fcs"
    )
    return Path(filepath)


def file_jpg_paradisi05() -> Path:
    """Return jpg file example.

    Originally from: https://upload.wikimedia.org/wikipedia/commons/2/28/Laminopathic_nuclei.jpg
    """  # noqa
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/Laminopathic_nuclei.jpg",
        "paradisi05_laminopathic_nuclei.jpg",
    )
    return Path(filepath)


def file_fastq() -> Path:
    """Mini mock fastq file."""
    with open("./input.fastq.gz", "w") as f:
        f.write("Mock fastq file.")
    return Path("./input.fastq.gz")


def file_bam() -> Path:
    """Mini mock bam file."""
    with open("./output.bam", "w") as f:
        f.write("Mock bam file.")
    return Path("./output.bam")


def file_mini_csv() -> Path:
    """Mini csv file."""
    filename = Path("./mini.csv")
    df = pd.DataFrame([1, 2, 3], columns=["test"])
    df.to_csv(filename, index=False)
    return filename


def dir_scrnaseq_cellranger() -> Path:
    """Directory with exemplary scrnaseq cellranger input and output."""
    filepath, _ = urlretrieve(
        "https://lamindb-test.s3.amazonaws.com/cellranger_run_001.zip"
    )
    from zipfile import ZipFile

    with ZipFile(filepath, "r") as zipObj:
        # Extract all the contents of zip file in current directory
        zipObj.extractall(path=".")

    return Path("cellranger_run_001")


def anndata_mouse_sc_lymph_node() -> ad.AnnData:
    """Mouse lymph node scRNA-seq dataset from EBI.

    Subsampled to 10k genes.

    From: https://www.ebi.ac.uk/arrayexpress/experiments/E-MTAB-8414/
    """
    filepath, _ = urlretrieve("https://lamindb-test.s3.amazonaws.com/E-MTAB-8414.h5ad")
    return ad.read(filepath)


def anndata_pbmc68k_reduced() -> ad.AnnData:
    """Modified from scanpy.datasets.pbmc68k_reduced()."""
    import scanpy as sc

    pbmc68k = sc.datasets.pbmc68k_reduced()
    pbmc68k.obs.rename(columns={"bulk_labels": "cell_type"}, inplace=True)
    pbmc68k.obs["cell_type"] = pbmc68k.obs["cell_type"].cat.rename_categories(
        {"Dendritic": "Dendritic cells", "CD14+ Monocyte": "CD14+ Monocytes"}
    )
    del pbmc68k.obs["G2M_score"]
    del pbmc68k.obs["S_score"]
    del pbmc68k.obs["phase"]
    del pbmc68k.obs["n_counts"]
    del pbmc68k.var["dispersions"]
    del pbmc68k.var["dispersions_norm"]
    del pbmc68k.var["means"]
    del pbmc68k.uns["rank_genes_groups"]
    del pbmc68k.uns["bulk_labels_colors"]
    del pbmc68k.raw
    sc.pp.subsample(pbmc68k, fraction=0.1, random_state=123)
    return pbmc68k


def anndata_pbmc3k_processed() -> ad.AnnData:
    """Modified from scanpy.pbmc3k_processed()."""
    import scanpy as sc

    pbmc3k = sc.datasets.pbmc3k_processed()
    pbmc3k.obs.rename(columns={"louvain": "cell_type"}, inplace=True)
    return pbmc3k


def anndata_human_immune_cells() -> ad.AnnData:
    """Cross-tissue immune cell analysis reveals tissue-specific features in humans.

    From: https://cellxgene.cziscience.com/collections/62ef75e4-cbea-454e-a0ce-998ec40223d3  # noqa
    Dataset: Global

    To reproduce the subsample:
    >>> adata = sc.read('Global.h5ad')
    >>> adata.obs = adata.obs[['donor_id', 'tissue', 'cell_type', 'assay', 'tissue_ontology_term_id', 'cell_type_ontology_term_id', 'assay_ontology_term_id']].copy() . # noqa
    >>> sc.pp.subsample(adata, fraction=0.005)
    >>> del adata.uns["development_stage_ontology_term_id_colors"]
    >>> del adata.uns["sex_ontology_term_id_colors"]
    >>> sc.write('human_immune.h5ad', adata)
    """
    filepath, _ = urlretrieve("https://lamindb-test.s3.amazonaws.com/human_immune.h5ad")
    return ad.read(filepath)


def anndata_with_obs():
    """Create a mini anndata with cell_type, disease and tissue."""
    import anndata as ad
    import bionty as bt

    celltypes = ["T cell", "hematopoietic stem cell", "hepatocyte", "my new cell type"]
    celltype_ids = ["CL:0000084", "CL:0000037", "CL:0000182", ""]
    diseases = [
        "chronic kidney disease",
        "liver lymphoma",
        "cardiac ventricle disorder",
        "Alzheimer disease",
    ]
    tissues = ["kidney", "liver", "heart", "brain"]
    df = pd.DataFrame()
    df["cell_type"] = celltypes * 10
    df["cell_type_id"] = celltype_ids * 10
    df["tissue"] = tissues * 10
    df["disease"] = diseases * 10
    df.index = "obs" + df.index.astype(str)

    adata = ad.AnnData(X=np.zeros(shape=(40, 100), dtype=np.float32), obs=df)
    adata.var.index = bt.Gene().df().head(100)["ensembl_gene_id"].values

    return adata


def generate_cell_ranger_files(
    sample_name: str, basedir: Union[str, Path] = "./", output_only: bool = True
):
    """Generate mock cell ranger outputs.

    Args:
        sample_name: name of the sample
        basedir: run directory
        output_only: only generate output files

    """
    basedir = Path(basedir)

    if not output_only:
        fastqdir = basedir / "fastq"
        fastqdir.mkdir(parents=True, exist_ok=True)
        fastqfile1 = fastqdir / f"{sample_name}_R1_001.fastq.gz"
        with open(fastqfile1, "w") as f:
            f.write(f"{id.base62(n_char=6)}")
        fastqfile2 = fastqdir / f"{sample_name}_R2_001.fastq.gz"
        fastqfile2.touch(exist_ok=True)
        with open(fastqfile2, "w") as f:
            f.write(f"{id.base62(n_char=6)}")

    sampledir = basedir / f"{sample_name}"
    for folder in ["raw_feature_bc_matrix", "filtered_feature_bc_matrix", "analysis"]:
        filedir = sampledir / folder
        filedir.mkdir(parents=True, exist_ok=True)

    for filename in [
        "web_summary.html",
        "metrics_summary.csv",
        "possorted_genome_bam.bam",
        "possorted_genome_bam.bam.bai",
        "molecule_info.h5",
        "cloupe.cloupe",
        "raw_feature_bc_matrix.h5",
        "raw_feature_bc_matrix/barcodes.tsv.gz",
        "raw_feature_bc_matrix/features.tsv.gz",
        "raw_feature_bc_matrix/matrix.mtx.gz",
        "filtered_feature_bc_matrix.h5",
        "filtered_feature_bc_matrix/barcodes.tsv.gz",
        "filtered_feature_bc_matrix/features.tsv.gz",
        "filtered_feature_bc_matrix/matrix.mtx.gz",
        "analysis/analysis.csv",
    ]:
        file = sampledir / filename
        with open(file, "w") as f:
            f.write(f"{id.base62(n_char=6)}")

    return sampledir


# def schmidt22_crispra_gws_IFNG() -> Path:
#     """CRISPRi screen dataset of Schmidt22.

#     Originally from: https://zenodo.org/record/5784651
#     """
#     filepath, _ = urlretrieve(
#         "https://lamindb-test.s3.amazonaws.com/schmidt22-crispra-gws-IFNG.csv",
#         "schmidt22-crispra-gws-IFNG.csv",
#     )
#     return Path(filepath)


# def schmidt22_perturbseq() -> Path:
#     """Perturb-seq dataset of Schmidt22.

#     Subsampled and converted to h5ad from R file: https://zenodo.org/record/5784651

#     To reproduce the subsample:
#     >>> adata = sc.read('HuTcellsCRISPRaPerturbSeq_Re-stimulated.h5ad')
#     >>> adata.obs = adata.obs[['cluster_name']]
#     >>> del adata.obsp
#     >>> del adata.var['features']
#     >>> del adata.obsm['X_pca']
#     >>> del adata.uns
#     >>> del adata.raw
#     >>> del adata.varm
#     >>> adata.obs = adata.obs.reset_index()
#     >>> del adata.obs['index']
#     >>> sc.pp.subsample(adata, 0.03)
#     >>> adata.write('schmidt22_perturbseq.h5ad')
#     """
#     filepath, _ = urlretrieve(
#         "https://lamindb-test.s3.amazonaws.com/schmidt22_perturbseq.h5ad",
#         "schmidt22_perturbseq.h5ad",
#     )
#     return Path(filepath)


# def dir_scrnaseq_cellranger_schmidt22() -> Path:
#     """BFXoutput directory containing Schmidt22_perturbseq."""
#     filepath, _ = urlretrieve(
#         "https://lamindb-test.s3.amazonaws.com/scrnaseq-cellranger-schmidt22.zip",
#     )
#     from zipfile import ZipFile

#     with ZipFile(filepath, "r") as zipObj:
#         # Extract all the contents of zip file in current directory
#         zipObj.extractall(path=".")

#     return Path("scrnaseq-cellranger-schmidt22")



from pathlib import Path, PurePath
from typing import Any, Optional, Tuple, Union

import lndb
import pandas as pd
from anndata import AnnData
from appdirs import AppDirs
from lamin_logger import logger
from lndb_storage import UPath
from lndb_storage.object import infer_suffix, size_adata, write_to_file
from lnschema_core import File, Run

from lamindb._features import get_features
from lamindb._settings import settings
from lamindb.dev.db._select import select
from lamindb.dev.hashing import hash_file

DIRS = AppDirs("lamindb", "laminlabs")

NO_NAME_ERROR = """\
Pass a name or key in `ln.File(...)` when ingesting in-memory data.
"""

NO_SOURCE_HINT = """\
Consider linking a run using the `run` argument, e.g., by calling ln.track().
"""


def serialize(
    data: Union[Path, UPath, str, pd.DataFrame, AnnData],
    name,
    format,
    key: Optional[str] = None,
) -> Tuple[Any, Union[Path, UPath], str, str]:
    """Serialize a data object that's provided as file or in memory."""
    # Convert str to either Path or UPath
    if isinstance(data, (str, Path, UPath)):
        filepath = UPath(data)  # returns Path for local
        try:  # check if file exists
            if not filepath.exists():
                raise FileNotFoundError
        except PermissionError:  # we will setup permissions later
            pass
        if isinstance(filepath, UPath):
            new_storage = list(filepath.parents)[-1]
            if not get_check_path_in_storage(filepath):
                raise ValueError(
                    "Currently do not support moving cloud data across buckets."
                    " Configure storage to point to your cloud bucket:\n"
                    f" `ln.setup.set.storage({new_storage})` or `lamin set --storage"
                    f" {new_storage}`"
                )
            root = lndb.settings.storage.root
            if isinstance(root, UPath):
                filepath = UPath(
                    filepath, **root._kwargs
                )  # inherit fsspec kwargs from root
        memory_rep = None
        if key is None:
            name = filepath.name
        else:
            name = PurePath(key).name
        # also see tests/test_file_hashing.py
        suffix = "".join(filepath.suffixes)
    # For now, in-memory objects are always saved to local_filepath first
    # This behavior will change in the future
    elif isinstance(data, (pd.DataFrame, AnnData)):
        if name is None and key is None:
            raise ValueError(NO_NAME_ERROR)
        if name is None:
            name = PurePath(key).name
        memory_rep = data
        suffix = infer_suffix(data, format)
        # the following filepath is always local
        if lndb.settings.storage.cache_dir is not None:
            filepath = lndb.settings.storage.cache_dir / name
        else:
            # this should likely be added to lndb.settings.storage
            cache_dir = Path(DIRS.user_cache_dir)
            cache_dir.mkdir(parents=True, exist_ok=True)
            filepath = cache_dir / name
        if filepath.suffixes == []:
            filepath = filepath.with_suffix(suffix)
        if suffix != ".zarr":
            write_to_file(data, filepath)
    else:
        raise NotImplementedError("Recording not yet implemented for this type.")
    return memory_rep, filepath, name, suffix


def get_hash(local_filepath, suffix, check_hash: bool = True):
    if suffix != ".zarr":  # if not streamed
        hash = hash_file(local_filepath)
        if not check_hash:
            return hash
        result = select(File, hash=hash).all()
        if len(result) > 0:
            msg = f"A file with same hash is already in the DB: {result}"
            if settings.error_on_file_hash_exists:
                hint = (
                    "💡 You can make this error a warning:\n"
                    "    ln.settings.error_on_file_hash_exists = False"
                )
                raise RuntimeError(f"{msg}\n{hint}")
            else:
                logger.warning(msg)
    else:
        hash = None
    return hash


def get_run(run: Optional[Run]) -> Optional[Run]:
    if run is None:
        from ._context import context

        run = context.run
        if run is None:
            logger.warning("No run & transform get linked to this file.")
            logger.hint(NO_SOURCE_HINT)
    # the following ensures that queried objects (within __init__)
    # behave like queried objects, only example right now: Run
    if hasattr(run, "_ln_identity_key") and run._ln_identity_key is not None:  # type: ignore  # noqa
        run._sa_instance_state.key = run._ln_identity_key  # type: ignore
    return run


def get_path_size_hash(
    filepath: Union[Path, UPath],
    memory_rep: Optional[Union[pd.DataFrame, AnnData]],
    suffix: str,
    check_hash: bool = True,
):
    cloudpath = None
    localpath = None

    if suffix == ".zarr":
        if memory_rep is not None:
            size = size_adata(memory_rep)
        else:
            if isinstance(filepath, UPath):
                cloudpath = filepath
                # todo: properly calculate size
                size = 0
            else:
                localpath = filepath
                size = sum(
                    f.stat().st_size for f in filepath.rglob("*") if f.is_file()  # type: ignore # noqa
                )
        hash = None
    else:
        if isinstance(filepath, UPath):
            try:
                size = filepath.stat()["size"]
            # here trying to fix access issue with new s3 buckets
            except Exception as e:
                if filepath._url.scheme == "s3":
                    filepath = UPath(filepath, cache_regions=True)
                    size = filepath.stat()["size"]
                else:
                    raise e
            cloudpath = filepath
            hash = None
        else:
            size = filepath.stat().st_size  # type: ignore
            localpath = filepath
            hash = get_hash(filepath, suffix, check_hash=check_hash)

    return localpath, cloudpath, size, hash


def get_check_path_in_storage(
    filepath: Union[Path, UPath], *, root: Optional[Union[Path, UPath]] = None
) -> bool:
    assert isinstance(filepath, Path)
    if root is None:
        root = lndb.settings.storage.root
    # the following comparisons can fail if types aren't comparable
    if isinstance(filepath, UPath) and isinstance(root, UPath):
        # the following tests equivalency of two UPath objects
        # via string representations; otherwise
        # S3Path('s3://lndb-storage/') and S3Path('s3://lamindb-ci/')
        # test as equivalent
        return list(filepath.parents)[-1].as_posix() == root.as_posix()
    elif not isinstance(filepath, UPath) and not isinstance(root, UPath):
        return root in filepath.resolve().parents
    else:
        return False


def get_relative_path_to_directory(
    path: Union[PurePath, Path, UPath], directory: Union[PurePath, Path, UPath]
) -> Union[PurePath, Path]:
    if isinstance(directory, UPath):
        # UPath.relative_to() is not behaving as it should (2023-04-07)
        relpath = PurePath(path.as_posix().replace(directory.as_posix(), ""))
    elif isinstance(directory, Path):
        relpath = path.resolve().relative_to(directory.resolve())  # type: ignore
    elif isinstance(directory, PurePath):
        relpath = path.relative_to(directory)
    else:
        raise TypeError("directory not of type Path or UPath")
    return relpath


def get_relative_path_to_root(
    path: Union[Path, UPath], *, root: Optional[Union[Path, UPath]] = None
) -> Union[PurePath, Path]:
    """Relative path to the storage root path."""
    if root is None:
        root = lndb.settings.storage.root
    return get_relative_path_to_directory(path, root)


# expose to user via ln.File
def get_file_kwargs_from_data(
    data: Union[Path, UPath, str, pd.DataFrame, AnnData],
    *,
    name: Optional[str] = None,
    key: Optional[str] = None,
    run: Optional[Run] = None,
    format: Optional[str] = None,
):
    run = get_run(run)
    memory_rep, filepath, safe_name, suffix = serialize(data, name, format, key)
    # the following will return a localpath that is not None if filepath is local
    # it will return a cloudpath that is not None if filepath is on the cloud
    local_filepath, cloud_filepath, size, hash = get_path_size_hash(
        filepath, memory_rep, suffix
    )
    check_path_in_storage = get_check_path_in_storage(filepath)
    # if we pass a file, no storage key, and path is already in storage,
    # then use the existing relative path within the storage location
    # as storage key
    if memory_rep is None and key is None and check_path_in_storage:
        key = get_relative_path_to_root(path=filepath).as_posix()

    kwargs = dict(
        name=safe_name,
        suffix=suffix,
        hash=hash,
        key=key,
        size=size,
        storage_id=lndb.settings.storage.id,
        # passing both the id and the object
        # to make them both available immediately
        # after object creation
        run_id=run.id if run is not None else None,
        run=run,
    )
    privates = dict(
        local_filepath=local_filepath,
        cloud_filepath=cloud_filepath,
        memory_rep=memory_rep,
        check_path_in_storage=check_path_in_storage,
    )

    return kwargs, privates


# expose to user via ln.Features
def get_features_from_data(
    data: Union[Path, UPath, str, pd.DataFrame, AnnData],
    reference: Any,
    format: Optional[str] = None,
    **curate_kwargs,
):
    # always convert to bionty.{entity}
    reference = reference.bionty if hasattr(reference, "bionty") else reference

    memory_rep, filepath, _, suffix = serialize(data, "features", format, key=None)
    localpath, cloudpath, _, _ = get_path_size_hash(
        filepath, memory_rep, suffix, check_hash=False
    )

    file_privates = dict(
        _local_filepath=localpath,
        _cloud_filepath=cloudpath,
        _memory_rep=memory_rep,
    )
    return get_features(file_privates, reference, **curate_kwargs)



"""v0.26.2.

Revision ID: 8280855a5064
Revises: ff3b5b3ec913
Create Date: 2023-02-14 22:21:20.501993

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8280855a5064"
down_revision = "ff3b5b3ec913"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema = "core.", None
    else:
        prefix, schema = "", "core"

    with op.batch_alter_table(f"{prefix}run", schema=schema) as batch_op:
        batch_op.add_column(sa.Column("external_id", sqlmodel.sql.sqltypes.AutoString(), nullable=True))
        if sqlite:
            batch_op.create_index(batch_op.f("ix_core.run_external_id"), ["external_id"], unique=False)
        else:
            batch_op.create_index(batch_op.f("ix_core_run_external_id"), ["external_id"], unique=False)


def downgrade() -> None:
    pass



from typing import Union

import erdiagram
import pydot
import sqlalchemy as sql
from lndb import settings
from sqlalchemy import Column, Table


def graph() -> pydot.Dot:
    """Get diagram of entity relationships as `pydot.Dot` graph object.

    It uses `erdiagram.create_schema_graph`
    """
    metadata = get_db_metadata()
    return erdiagram.create_schema_graph(
        metadata=metadata,
        show_datatypes=False,
        show_indexes=False,
        rankdir="TB",
        concentrate=True,
    )


def view():
    """View diagram of entity relationships.

    It displays :func:`~lamindb.schema.graph`.
    """
    erdiagram.view(graph())


def list_tables():
    """Return all entities."""
    metadata = get_db_metadata()
    table_names = [table.name for table in metadata.sorted_tables]
    return table_names


def get_db_metadata():
    engine = settings.instance.engine
    metadata = sql.MetaData(bind=engine)
    metadata.reflect()
    return metadata


def get_db_metadata_as_dict():
    metadata = get_db_metadata()
    return {
        "key": get_db_name(),
        "tables": {
            table_name: get_table_metadata_as_dict(table)
            for table_name, table in metadata.tables.items()
        },
    }


def get_table_metadata_as_dict(table: Table):
    return {
        "key": table.key,
        "primary_keys": table.primary_key.columns.keys(),
        "foreign_keys": get_foreign_keys_as_tuples(table),
        "columns": {
            column_name: get_column_metadata_as_dict(column)
            for column_name, column in table.columns.items()
        },
    }


def get_column_metadata_as_dict(column: Column):
    return {
        "key": column.key,
        "type": str(column.type),
        "primary_key": column.primary_key,
        "foreign_keys": get_foreign_keys_as_tuples(column),
        "nullable": column.nullable,
        "default": column.default,
    }


def get_foreign_keys_as_tuples(object: Union[Table, Column]):
    return [(fk.column.table.key, fk.column.key) for fk in object.foreign_keys]


def get_db_name() -> str:
    engine = settings.instance.engine
    return engine.url.database


def get_table_object(table_name: str):
    metadata = get_db_metadata()
    return metadata.tables[table_name]



import lndb as _lndb
from lndb import *
from lndb import settings

from . import dev

__doc__ = _lndb.__doc__.replace("lndb", "lamindb.setup")
settings.__doc__ = settings.__doc__.replace("lndb", "lamindb.setup")
